from unittest import mock

from api_fhir_r4.converters import PatientConverter

from fhir.resources.patient import Patient
from api_fhir_r4.tests import PatientTestMixin
from api_fhir_r4.tests.mixin import ConvertToFhirTestMixin, ConvertToImisTestMixin, ConvertJsonToFhirTestMixin


class PatientConverterTestCase(PatientTestMixin,
                               ConvertToFhirTestMixin,
                               ConvertToImisTestMixin,
                               ConvertJsonToFhirTestMixin):
    converter = PatientConverter
    fhir_resource = Patient
    json_repr = 'test/test_patient.json'

    @mock.patch('insuree.models.Gender.objects')
    def test_to_imis_obj(self, mock_gender):
        mock_gender.get.return_value = self._TEST_GENDER
        super(PatientConverterTestCase, self).test_to_imis_obj()




from rest_framework import status


class FhirApiCreateTestMixin(object):

    @property
    def base_url(self):
        raise NotImplementedError()

    @property
    def _test_request_data(self):
        raise NotImplementedError()

    def login(self):
        raise NotImplementedError()

    def test_post_should_create_correctly(self):
        self.login()
        self.create_dependencies()
        response = self.client.post(self.base_url, data=self._test_request_data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertIsNotNone(response.content)

    def create_dependencies(self):
        pass



from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiReadTestMixin
from rest_framework.test import APITestCase


class OrganisationAPITests(GenericFhirAPITestMixin, FhirApiReadTestMixin, APITestCase):
    base_url = GeneralConfiguration.get_base_url() + 'Organization/'
    _test_json_path = None

    def setUp(self):
        super(OrganisationAPITests, self).setUp()



from api_fhir_r4.configurations import IdentifierConfiguration


class R4IdentifierConfig(IdentifierConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_identifier_type_config = cfg['R4_fhir_identifier_type_config']

    @classmethod
    def get_fhir_identifier_type_system(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config")\
            .get('system', 'https://openimis.github.io/openimis_fhir_r4_ig/CodeSystem/openimis-identifiers')

    @classmethod
    def get_fhir_acsn_type_code(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_db_id_type', "ACSN")

    @classmethod
    def get_fhir_id_type_code(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_db_id_type', "ACSN")

    @classmethod
    def get_fhir_chfid_type_code(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_chfid_type', "SB")

    @classmethod
    def get_fhir_passport_type_code(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_passport_type', "PPN")

    @classmethod
    def get_fhir_facility_id_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_facility_id_type', "FI")

    @classmethod
    def get_fhir_claim_admin_code_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_claim_admin_code_type', "FILL")

    @classmethod
    def get_fhir_claim_code_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_claim_code_type', "Code")

    @classmethod
    def get_fhir_uuid_type_code(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_db_uuid_type', "UUID")

    @classmethod
    def get_fhir_generic_type_code(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_generic_code', "Code")

    @classmethod
    def get_fhir_location_code_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_location_code_type', "Code")

    @classmethod
    def get_fhir_diagnosis_code_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_diagnosis_code_type', "DC")

    @classmethod
    def get_fhir_item_code_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_item_code_type', "IC")

    @classmethod
    def get_fhir_service_code_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get('fhir_code_for_imis_service_code_type', "SC")



from django.utils.translation import gettext as _
from fhir.resources.identifier import Identifier
from fhir.resources.address import Address
from fhir.resources.extension import Extension
from fhir.resources.humanname import HumanName
from fhir.resources.organization import Organization

from api_fhir_r4.models.imisModelEnums import ContactPointSystem
from api_fhir_r4.configurations import R4IdentifierConfig, R4OrganisationConfig, GeneralConfiguration
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin


class InsuranceOrganisationConverter(BaseFHIRConverter):

    @classmethod
    def to_fhir_obj(cls, imis_organisation, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_organisation = Organization()
        cls.build_fhir_id(fhir_organisation, imis_organisation)
        cls.build_fhir_identifier_id(
            fhir_organisation,
            imis_organisation,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            cls.get_fhir_code_identifier_type()
        )
        cls.build_fhir_name(fhir_organisation, imis_organisation)
        cls.build_fhir_type(fhir_organisation, imis_organisation)
        cls.build_fhir_contact(fhir_organisation, imis_organisation)
        cls.build_fhir_telecom(fhir_organisation, imis_organisation)
        cls.build_fhir_address_organisation(fhir_organisation, imis_organisation)
        return fhir_organisation

    @classmethod
    def to_imis_obj(cls, fhir_organisation, audit_user_id):
        raise NotImplementedError(
            _('Insurance Organisation to_imis_obj() not implemented.')
        )

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def get_fhir_resource_type(cls):
        return Organization

    @classmethod
    def build_fhir_id(cls, fhir_organisation, imis_organisation):
        fhir_organisation.id = f"{imis_organisation['id']}"

    @classmethod
    def build_fhir_identifier_id(cls, fhir_organisation, imis_organisation, type_system, type_code):
        fhir_organisation.identifier = []
        identifier = Identifier.construct()
        identifier.type = cls.build_codeable_concept(type_code, type_system)
        identifier.value = imis_organisation["code"]
        fhir_organisation.identifier.append(identifier)

    @classmethod
    def build_fhir_type(cls, fhir_organisation, imis_organisation):
        fhir_organisation.type = [cls.build_codeable_concept(
            code=imis_organisation["type"],
            system=R4OrganisationConfig.get_fhir_ph_organisation_type_system()
        )]

    @classmethod
    def build_fhir_name(cls, fhir_organisation, imis_organisation):
        fhir_organisation.name = imis_organisation["name"]

    @classmethod
    def build_fhir_telecom(cls, fhir_organisation, imis_organisation):
        fhir_organisation.telecom = []
        fhir_organisation.telecom.append(cls.build_fhir_contact_point(
            system=ContactPointSystem.EMAIL,
            value=imis_organisation["email"]))

        fhir_organisation.telecom.append(cls.build_fhir_contact_point(
            system=ContactPointSystem.FAX,
            value=imis_organisation["fax"]))

        fhir_organisation.telecom.append(cls.build_fhir_contact_point(
            system=ContactPointSystem.PHONE,
            value=imis_organisation["phone"]))

    @classmethod
    def build_fhir_contact(cls, fhir_organisation, imis_organisation):
        fhir_organisation.contact = []

        name = HumanName.construct()
        name.text = imis_organisation['contact_name']

        purpose = cls.build_codeable_concept(
            code="ADMIN",
            system=R4OrganisationConfig.get_fhir_ph_organisation_contactentity_type()
        )
        if len(purpose.coding) == 1:
            purpose.coding[0].display = _("Administrative")

        fhir_organisation.contact.append({'name': name, 'purpose': purpose})

    @classmethod
    def build_fhir_address_organisation(cls, fhir_organization, imis_organisation):
        addresses = []

        address = Address.construct()

        # municipality extension
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-municipality"
        extension.valueString = imis_organisation["municipality"]
        address.extension = [extension]

        address.line = [imis_organisation["line"]]
        address.state = imis_organisation["state"]
        address.district = imis_organisation["district"]
        address.city = imis_organisation["city"]

        addresses.append(address)
        fhir_organization.address = addresses



import copy

from location.models import Location, HealthFacility

from api_fhir_r4.converters import LocationSiteConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


class LocationSiteSerializer(BaseFHIRSerializer):
    fhirConverter = LocationSiteConverter()

    def create(self, validated_data):
        copied_data = copy.deepcopy(validated_data)
        del copied_data['_state']
        return HealthFacility.objects.create(**copied_data)

    def update(self, instance, validated_data):
        # TODO legalForm isn't covered because that value is missing in the model
        # TODO LocationId isn't covered because that value is missing in the model
        # TODO offline isn't covered in the current version of API
        # TODO care_type isn't covered in the current version of API
        instance.code = validated_data.get('code', instance.code)
        instance.name = validated_data.get('name', instance.name)
        instance.type = validated_data.get('type', instance.type)
        instance.partOf = validated_data.get('partOf', instance.partOf)
        instance.audit_user_id = self.get_audit_user_id()
        instance.save()
        return instance



import logging
from policy.services import EligibilityRequest, EligibilityService, EligibilityResponse, StoredProcEligibilityService
from policy.services import ByInsureeRequest, ByInsureeService, ByInsureeResponse
from api_fhir_r4.converters import CoverageEligibilityRequestConverter
from api_fhir_r4.serializers import BaseFHIRSerializer
from django.http.response import HttpResponseBase
from fhir.resources.fhirabstractmodel import FHIRAbstractModel
from api_fhir_r4.converters import OperationOutcomeConverter


class CoverageEligibilityRequestSerializer(BaseFHIRSerializer):

    fhirConverter = CoverageEligibilityRequestConverter()
    logger = logging.getLogger(__name__)

    def to_representation(self, obj):
        if isinstance(obj, HttpResponseBase):
            return OperationOutcomeConverter.to_fhir_obj(obj).dict()
        elif isinstance(obj, FHIRAbstractModel):
            return obj.dict()
        return CoverageEligibilityRequestConverter.to_fhir_obj(obj[0], obj[1]).dict()

    def create(self, validated_data):
        eligibility_request_sp = EligibilityRequest(
            chf_id=validated_data.get('chf_id'),
            service_code=validated_data.get('service_code'),
            item_code=validated_data.get('item_code')
        )
        eligibility_request = ByInsureeRequest(chf_id=validated_data.get('chf_id'))
        request = self.context.get("request")
        try:
            response = ByInsureeService(request.user).request(eligibility_request)
        except TypeError:
            self.logger.warning('The insuree with chfid `{}` is not connected with policy. '
                                'The default eligibility response will be used.'
                                .format(validated_data.get('chfid')))
            response = self.create_default_eligibility_response()
        output_response = [response, eligibility_request_sp]
        return output_response

    def create_default_eligibility_response(self):
        return ByInsureeResponse(
            by_insuree_request=None,
            items=[]
        )



from fhir.resources.practitionerrole import PractitionerRole
from fhir.resources.extension import Extension
from fhir.resources.identifier import Identifier
from fhir.resources.reference import Reference
from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import EnrolmentOfficerPractitionerConverter, EnrolmentOfficerPractitionerRoleConverter
from api_fhir_r4.tests import GenericTestMixin, EnrolmentOfficerPractitionerTestMixin, LocationTestMixin
from core.models import Officer


class EnrolmentOfficerPractitionerRoleTestMixin(GenericTestMixin):

    _TEST_OFFICER = None
    _TEST_LOCATION = None
    _TEST_SUBSTITUTION_OFFICER = None
    _TEST_SUBSTITUTION_OFFICER_UUID = "f4bf924a-e2c9-46dc-9fa3-d54a1a67ea86"
    _TEST_SUBSTITUTION_OFFICER_CODE = "EOTESB"
    _TEST_PRACTITIONER_REFERENCE = None
    _TEST_LOCATION_REFERENCE = None
    _TEST_LOCATION_UUID = "041890f3-d90b-46ca-8b25-56c3f8d60615"

    def setUp(self):
        self._TEST_OFFICER = EnrolmentOfficerPractitionerTestMixin().create_test_imis_instance()
        self._TEST_SUBSTITUTION_OFFICER = self.create_test_subsitution_officer()
        self._TEST_PRACTITIONER_REFERENCE = "Practitioner/" + self._TEST_OFFICER.uuid
        self._TEST_LOCATION = self.create_test_location()
        self._TEST_LOCATION_REFERENCE = "Location/" + self._TEST_LOCATION.uuid

    def create_test_subsitution_officer(self):
        officer = Officer.objects.create(
            **{
                "uuid": self._TEST_SUBSTITUTION_OFFICER_UUID,
                "code": self._TEST_SUBSTITUTION_OFFICER_CODE,
                "last_name": "Officer",
                "other_names": "Test",
                "validity_to": None,
                "audit_user_id": -1,
            }
        )
        return officer

    def create_test_location(self):
        location = LocationTestMixin().create_test_imis_instance()
        location.uuid = self._TEST_LOCATION_UUID
        location.save()
        return location

    def create_test_imis_instance(self):
        self._TEST_OFFICER.location = self._TEST_LOCATION
        self._TEST_OFFICER.substitution_officer = self._TEST_SUBSTITUTION_OFFICER
        self._TEST_OFFICER.audit_user_id = -1
        self._TEST_OFFICER.save()
        return self._TEST_OFFICER

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_OFFICER.substitution_officer.code, imis_obj.substitution_officer.code)
        self.assertEqual(self._TEST_LOCATION.code, imis_obj.location.code)

    def create_test_fhir_instance(self):
        self.create_test_imis_instance()
        fhir_practitioner_role = PractitionerRole.construct()
        identifiers = []
        code = EnrolmentOfficerPractitionerRoleConverter.build_fhir_identifier(
            self._TEST_OFFICER.code,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code()
        )
        identifiers.append(code)
        location_reference = Reference.construct()
        location_reference.reference = self._TEST_LOCATION_REFERENCE
        fhir_practitioner_role.location = [location_reference]
        practitioner_reference = Reference.construct()
        practitioner_reference.reference = self._TEST_PRACTITIONER_REFERENCE
        fhir_practitioner_role.practitioner = practitioner_reference
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/practitioner-role-substitution-reference"
        reference = EnrolmentOfficerPractitionerConverter.build_fhir_resource_reference(
            self._TEST_SUBSTITUTION_OFFICER,
        )
        extension.valueReference = reference
        fhir_practitioner_role.extension = [extension]
        return fhir_practitioner_role

    def verify_fhir_instance(self, fhir_obj):
        self.assertIn(self._TEST_LOCATION_REFERENCE, fhir_obj.location[0].reference)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = EnrolmentOfficerPractitionerRoleConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_generic_type_code():
                self.assertEqual(self._TEST_OFFICER.code, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(self._TEST_OFFICER.uuid, identifier.value)
        self.assertIn(self._TEST_PRACTITIONER_REFERENCE, fhir_obj.practitioner.reference)
        self.assertIn(self._TEST_SUBSTITUTION_OFFICER.uuid, fhir_obj.extension[0].valueReference.reference.split('Practitioner/')[1])
        self.assertEqual(1, len(fhir_obj.code))
        self.assertEqual(1, len(fhir_obj.code[0].coding))
        self.assertEqual("EO", fhir_obj.code[0].coding[0].code)
        self.assertEqual("Enrolment Officer",  fhir_obj.code[0].coding[0].display)



from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiReadTestMixin
from api_fhir_r4.configurations import GeneralConfiguration
from rest_framework.test import APITestCase


class PractitionerRoleAPITests(GenericFhirAPITestMixin, FhirApiReadTestMixin, APITestCase):

    base_url = GeneralConfiguration.get_base_url()+'PractitionerRole/'
    _test_json_path = "/test/test_claimAdminPractitionerRole.json"

    def setUp(self):
        super(PractitionerRoleAPITests, self).setUp()



import copy

from insuree.apps import InsureeConfig
from insuree.models import Insuree, Family
from insuree.gql_mutations import create_file

from api_fhir_r4.converters import PatientConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers import BaseFHIRSerializer
from insuree.services import InsureeService


class PatientSerializer(BaseFHIRSerializer):
    fhirConverter = PatientConverter()

    def create(self, validated_data):
        self._validate_data(validated_data.get('chf_id'))
        copied_data = self._clean_data(copy.deepcopy(validated_data))
        obj = InsureeService(self.context.get("request").user)\
            .create_or_update(copied_data)
        # create photo as a file to specified configured path
        if InsureeConfig.insuree_photos_root_path:
            create_file(date=obj.photo.date, insuree_id=obj.id, photo_bin=obj.photo.photo)

        if copied_data['head']:
            self._create_patient_family(obj, validated_data)
        return obj

    def update(self, instance, validated_data):
        request = self.context.get("request")
        user = request.user
        chf_id = validated_data.get('chf_id')
        if Insuree.objects.filter(chf_id=chf_id).count() == 0:
            raise FHIRException('No patients with following chfid `{}`'.format(chf_id))
        insuree = Insuree.objects.get(chf_id=chf_id, validity_to__isnull=True)
        validated_data["id"] = insuree.id
        validated_data["uuid"] = insuree.uuid
        del validated_data['_state']
        instance = InsureeService(user).create_or_update(validated_data)
        return instance

    def _validate_data(self, chf_id):
        if not chf_id:
            raise FHIRException("Provided patient without code.")

        if Insuree.objects.filter(chf_id=chf_id, validity_to__isnull=True).exists():
            raise FHIRException('Exists patient with following chfid `{}`'.format(chf_id))

    def _clean_data(self, validated_data):
        validated_data.pop('_state', None)
        validated_data.pop('family_address', None)
        validated_data.pop('family_location', None)
        return validated_data

    def _create_patient_family(self, obj, validated_data):
        audit_user_id = validated_data['audit_user_id']
        family_location = validated_data.get('family_location', None)
        family_address = validated_data.get('family_address', None)

        obj.family = Family.objects.create(
            location=family_location,
            address=family_address,
            head_insuree=obj,
            audit_user_id=audit_user_id
        )
        obj.save()



from api_fhir_r4.configurations import R4SubscriptionConfig
from core.apps import CoreConfig
from claim.apps import ClaimConfig
from insuree.apps import InsureeConfig
from location.apps import LocationConfig
from policy.apps import PolicyConfig
from policyholder.apps import PolicyholderConfig
from product.apps import ProductConfig
from medical.apps import MedicalConfig
from invoice.apps import InvoiceConfig
from rest_framework import exceptions
from rest_framework.permissions import DjangoModelPermissions


class FHIRApiPermissions(DjangoModelPermissions):
    permissions_get = []
    permissions_post = []
    permissions_put = []
    permissions_patch = []
    permissions_delete = []

    def __init__(self):
        self.perms_map['GET'] = self.permissions_get
        self.perms_map['POST'] = self.permissions_post
        self.perms_map['PUT'] = self.permissions_put
        self.perms_map['PATCH'] = self.permissions_patch
        self.perms_map['DELETE'] = self.permissions_delete

    def get_required_permissions(self, method, model_cls):
        if method not in self.perms_map:
            raise exceptions.MethodNotAllowed(method)

        return self.perms_map[method]


class FHIRApiClaimPermissions(FHIRApiPermissions):
    permissions_get = ClaimConfig.gql_query_claims_perms
    permissions_post = ClaimConfig.gql_mutation_create_claims_perms
    permissions_put = ClaimConfig.gql_mutation_update_claims_perms
    permissions_patch = ClaimConfig.gql_mutation_update_claims_perms
    permissions_delete = ClaimConfig.gql_mutation_delete_claims_perms


class FHIRApiCommunicationRequestPermissions(FHIRApiPermissions):
    permissions_get = ClaimConfig.gql_mutation_select_claim_feedback_perms
    permissions_post = ClaimConfig.gql_mutation_deliver_claim_feedback_perms
    permissions_put = ClaimConfig.gql_mutation_deliver_claim_feedback_perms
    permissions_patch = ClaimConfig.gql_mutation_deliver_claim_feedback_perms
    permissions_delete = ClaimConfig.gql_mutation_skip_claim_feedback_perms


class FHIRApiPractitionerPermissions(FHIRApiPermissions):
    permissions_get = CoreConfig.gql_query_claim_administrator_perms
    permissions_post = CoreConfig.gql_mutation_create_claim_administrator_perms
    permissions_put = CoreConfig.gql_mutation_update_claim_administrator_perms
    permissions_patch = CoreConfig.gql_mutation_update_claim_administrator_perms
    permissions_delete = CoreConfig.gql_mutation_delete_claim_administrator_perms


class FHIRApiPractitionerOfficerPermissions(FHIRApiPermissions):
    permissions_get = CoreConfig.gql_query_enrolment_officers_perms
    permissions_post = CoreConfig.gql_mutation_create_enrolment_officers_perms
    permissions_put = CoreConfig.gql_mutation_update_enrolment_officers_perms
    permissions_patch = CoreConfig.gql_mutation_update_enrolment_officers_perms
    permissions_delete = CoreConfig.gql_mutation_delete_enrolment_officers_perms


class FHIRApiCoverageEligibilityRequestPermissions(FHIRApiPermissions):
    permissions_get = PolicyConfig.gql_query_eligibilities_perms
    permissions_post = []
    permissions_put = []
    permissions_patch = []
    permissions_delete = []


class FHIRApiCoverageRequestPermissions(FHIRApiPermissions):
    permissions_get = PolicyConfig.gql_query_policies_by_insuree_perms
    permissions_post = []
    permissions_put = []
    permissions_patch = []
    permissions_delete = []


class FHIRApiHFPermissions(FHIRApiPermissions):
    permissions_get = LocationConfig.gql_query_health_facilities_perms
    permissions_post = LocationConfig.gql_mutation_create_health_facilities_perms
    permissions_put = LocationConfig.gql_mutation_create_health_facilities_perms
    permissions_patch = LocationConfig.gql_mutation_create_health_facilities_perms
    permissions_delete = LocationConfig.gql_mutation_delete_health_facilities_perms


class FHIRApiInsuranceOrganizationPermissions(FHIRApiPermissions):
    permissions_get = LocationConfig.gql_query_health_facilities_perms
    permissions_post = []
    permissions_put = []
    permissions_patch = []
    permissions_delete = []


class FHIRApiInsureePermissions(FHIRApiPermissions):
    permissions_get = InsureeConfig.gql_query_insurees_perms
    permissions_post = InsureeConfig.gql_mutation_create_insurees_perms
    permissions_put = InsureeConfig.gql_mutation_update_insurees_perms
    permissions_patch = InsureeConfig.gql_mutation_update_insurees_perms
    permissions_delete = InsureeConfig.gql_mutation_delete_insurees_perms


class FHIRApiMedicationPermissions(FHIRApiPermissions):
    permissions_get = MedicalConfig.gql_query_medical_items_perms
    permissions_post = MedicalConfig.gql_mutation_medical_items_add_perms
    permissions_put = MedicalConfig.gql_mutation_medical_items_update_perms
    permissions_patch = MedicalConfig.gql_mutation_medical_items_update_perms
    permissions_delete = MedicalConfig.gql_mutation_medical_items_delete_perms


class FHIRApiConditionPermissions(FHIRApiPermissions):
    permissions_get = MedicalConfig.gql_query_diagnosis_perms
    permissions_post = []
    permissions_put = []
    permissions_patch = []
    permissions_delete = []


class FHIRApiActivityDefinitionPermissions(FHIRApiPermissions):
    permissions_get = MedicalConfig.gql_query_medical_services_perms
    permissions_post = MedicalConfig.gql_mutation_medical_services_add_perms
    permissions_put = MedicalConfig.gql_mutation_medical_services_update_perms
    permissions_patch = MedicalConfig.gql_mutation_medical_services_update_perms
    permissions_delete = MedicalConfig.gql_mutation_medical_services_delete_perms


class FHIRApiHealthServicePermissions(FHIRApiPermissions):
    permissions_get = LocationConfig.gql_query_health_facilities_perms
    permissions_post = LocationConfig.gql_mutation_create_health_facilities_perms
    permissions_put = LocationConfig.gql_mutation_edit_health_facilities_perms
    permissions_patch = LocationConfig.gql_mutation_edit_health_facilities_perms
    permissions_delete = LocationConfig.gql_mutation_delete_health_facilities_perms


class FHIRApiGroupPermissions(FHIRApiPermissions):
    permissions_get = InsureeConfig.gql_query_families_perms
    permissions_post = InsureeConfig.gql_mutation_create_families_perms
    permissions_put = InsureeConfig.gql_mutation_update_families_perms
    permissions_patch = InsureeConfig.gql_mutation_update_families_perms
    permissions_delete = InsureeConfig.gql_mutation_delete_families_perms


class FHIRApiOrganizationPermissions(FHIRApiPermissions):
    permissions_get = PolicyholderConfig.gql_query_policyholder_perms
    permissions_post = PolicyholderConfig.gql_mutation_create_policyholder_perms
    permissions_put = PolicyholderConfig.gql_mutation_update_policyholder_perms
    permissions_patch = PolicyholderConfig.gql_mutation_update_policyholder_perms
    permissions_delete = PolicyholderConfig.gql_mutation_delete_policyholder_perms


class FHIRApiProductPermissions(FHIRApiPermissions):
    permissions_get = ProductConfig.gql_query_products_perms
    permissions_post = ProductConfig.gql_mutation_products_add_perms
    permissions_put = ProductConfig.gql_mutation_products_edit_perms
    permissions_patch = ProductConfig.gql_mutation_products_edit_perms
    permissions_delete = ProductConfig.gql_mutation_products_delete_perms


class FHIRApiInvoicePermissions(FHIRApiPermissions):
    permissions_get = InvoiceConfig.gql_invoice_search_perms
    permissions_post = InvoiceConfig.gql_invoice_create_perms
    permissions_put = InvoiceConfig.gql_invoice_update_perms
    permissions_patch = InvoiceConfig.gql_invoice_update_perms
    permissions_delete = InvoiceConfig.gql_invoice_delete_perms


class FHIRApiBillPermissions(FHIRApiPermissions):
    permissions_get = InvoiceConfig.gql_bill_search_perms
    permissions_post = InvoiceConfig.gql_bill_create_perms
    permissions_put = InvoiceConfig.gql_bill_update_perms
    permissions_patch = InvoiceConfig.gql_bill_update_perms
    permissions_delete = InvoiceConfig.gql_bill_delete_perms


class FHIRApiSubscriptionPermissions(FHIRApiPermissions):
    permissions_get = R4SubscriptionConfig.get_fhir_sub_search_perms()
    permissions_post = R4SubscriptionConfig.get_fhir_sub_create_perms()
    permissions_put = R4SubscriptionConfig.get_fhir_sub_update_perms()
    permissions_patch = R4SubscriptionConfig.get_fhir_sub_update_perms()
    permissions_delete = R4SubscriptionConfig.get_fhir_sub_delete_perms()



from api_fhir_r4.models import Subscription
from core.validation import BaseModelValidation


class SubscriptionValidation(BaseModelValidation):
    OBJECT_TYPE = Subscription

    @classmethod
    def validate_create(cls, user, **data):
        super().validate_create(user, **data)

    @classmethod
    def validate_update(cls, user, **data):
        super().validate_update(user, **data)

    @classmethod
    def validate_delete(cls, user, **data):
        super().validate_delete(user, **data)



import core

from django.db.models import Q
from django.utils.translation import gettext as _
from api_fhir_r4.configurations import GeneralConfiguration, R4CoverageConfig
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from api_fhir_r4.converters.patientConverter import PatientConverter
from api_fhir_r4.mapping.contractMapping import PayTypeMapping, ContractStatus, \
    ContractState
from fhir.resources.contract import Contract, ContractTermAssetValuedItem, \
    ContractTerm, ContractTermAsset, ContractTermOffer, ContractTermOfferParty
from fhir.resources.extension import Extension
from fhir.resources.money import Money
from fhir.resources.period import Period

from product.models import Product
from policy.models import Policy
from insuree.models import Insuree, InsureePolicy
from insuree.models import Family
from contribution.models import Premium
from core.models import Officer
from api_fhir_r4.utils import DbManagerUtils, TimeUtils


class ContractConverter(BaseFHIRConverter, ReferenceConverterMixin):
    @classmethod
    def to_fhir_obj(cls, imis_policy, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_contract = Contract.construct()
        cls.build_contract_identifier(fhir_contract, imis_policy)
        cls.build_contract_author(fhir_contract, imis_policy, reference_type)
        cls.build_contract_subject(fhir_contract, imis_policy, reference_type)
        cls.build_contract_scope(fhir_contract, imis_policy)
        contract_term = ContractTerm.construct()
        cls.build_contract_term_offer(contract_term, imis_policy, reference_type)
        contract_term_asset = ContractTermAsset.construct()
        cls.build_contract_asset_extension(contract_term_asset, imis_policy, reference_type)
        cls.build_contract_asset_type_reference(contract_term_asset, imis_policy, reference_type)
        cls.build_contract_valued_item_entity(contract_term_asset, imis_policy)
        cls.build_contract_asset_use_period(contract_term_asset, imis_policy)
        contract_term.asset = [contract_term_asset]
        fhir_contract.term = [contract_term]
        cls.build_contract_status(fhir_contract, imis_policy)
        cls.build_contract_state(fhir_contract, imis_policy)
        return fhir_contract

    @classmethod
    def to_imis_obj(cls,fhir_contract, audit_user_id):
        errors = []
        fhir_contract = Contract(**fhir_contract)
        imis_policy = Policy()
        imis_policy.audit_user_id = audit_user_id
        cls.build_imis_period(imis_policy, fhir_contract.term, errors)
        cls.build_imis_useperiod(imis_policy, fhir_contract.term, errors)
        cls.build_imis_status(fhir_contract,imis_policy, errors)
        cls.build_imis_author(fhir_contract, imis_policy, errors)
        cls.build_imis_subject(fhir_contract, imis_policy, errors)
        cls.build_imis_product(fhir_contract, imis_policy, errors)
        cls.build_imis_state(fhir_contract, imis_policy, errors)
        cls.build_imis_insurees(fhir_contract, imis_policy, errors)
        cls.build_imis_contributions(fhir_contract, imis_policy, errors)
        cls.check_errors(errors)
        return imis_policy

    @classmethod
    def get_reference_obj_uuid(cls, imis_policy: Policy):
        return imis_policy.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_policy: Policy):
        return imis_policy.id


    @classmethod
    def get_fhir_resource_type(cls):
        return Contract

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_policy_code = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Policy, code=imis_policy_code)

    @classmethod
    def build_contract_identifier(cls, fhir_contract, imis_policy):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_policy)
        fhir_contract.identifier = identifiers
        return fhir_contract

    @classmethod
    def build_all_identifiers(cls, identifiers, imis_object):
        # Coverage have only uuid coverage
        cls.build_fhir_uuid_identifier(identifiers, imis_object)
        cls.build_fhir_id_identifier(identifiers, imis_object)
        return identifiers

    @classmethod
    def build_contract_author(cls, fhir_contract, imis_policy, reference_type):
        author = cls.build_fhir_resource_reference(
            imis_policy.officer, "Practitioner", imis_policy.officer.code, reference_type=reference_type)
        fhir_contract.author = author

    @classmethod
    def build_contract_subject(cls, fhir_contract, imis_policy, reference_type):
        subject = cls.build_fhir_resource_reference(
            imis_policy.family, "Group", imis_policy.family.head_insuree.last_name, reference_type=reference_type)
        fhir_contract.subject = [subject]

    @classmethod
    def build_contract_scope(cls, fhir_contract, imis_policy):
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/contract-scope"
        fhir_contract.scope = cls.build_codeable_concept(code="informal", system=system)
        if len(fhir_contract.scope.coding) == 1:
            fhir_contract.scope.coding[0].display = _("Informal Sector")

    @classmethod
    def build_contract_asset_extension(cls, contract_term_asset, imis_policy, reference_type):
        cls.build_contract_asset_premium(contract_term_asset, imis_policy)

    @classmethod
    def build_contract_asset_premium(cls, contract_term_asset, imis_policy):
        asset_extensions = Extension.construct()
        asset_extensions.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/contract-premium"
        if Premium.objects.filter(policy=imis_policy, validity_to__isnull=True).count() > 0:
            imis_premium = Premium.objects.get(policy=imis_policy, validity_to__isnull=True)
            fhir_premium = cls.build_contract_asset_premium_extension(asset_extensions, imis_premium)
            if type(contract_term_asset.extension) is not list:
                contract_term_asset.extension = [fhir_premium]
            else:
                contract_term_asset.extension.append(fhir_premium)

    @classmethod
    def build_contract_asset_premium_extension(cls, asset_extensions, imis_premium):
        cls.build_premium_payer_ext(asset_extensions)
        cls.build_premium_category_ext(asset_extensions)
        cls.build_premium_amount_ext(asset_extensions, imis_premium)
        cls.build_premium_receipt_ext(asset_extensions, imis_premium)
        cls.build_premium_date_ext(asset_extensions, imis_premium)
        cls.build_premium_type_ext(asset_extensions, imis_premium)
        return asset_extensions

    @classmethod
    def build_premium_payer_ext(cls, asset_extensions):
        extension = Extension.construct()
        extension.url = "payer"
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/contract-premium-payer"
        extension.valueCodeableConcept = cls.build_codeable_concept(code="beneficiary", system=system)
        if len(extension.valueCodeableConcept.coding) == 1:
            extension.valueCodeableConcept.coding[0].display = _("Beneficiary")
        asset_extensions.extension = [extension]

    @classmethod
    def build_premium_category_ext(cls, asset_extensions):
        extension = Extension.construct()
        extension.url = "category"
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/contract-premium-category"
        extension.valueCodeableConcept = cls.build_codeable_concept(code="C", system=system)
        if len(extension.valueCodeableConcept.coding) == 1:
            extension.valueCodeableConcept.coding[0].display = _("Contribution and Others")
        asset_extensions.extension.append(extension)

    @classmethod
    def build_premium_amount_ext(cls, asset_extensions, imis_premium):
        # get the currency defined in configs from core module
        if hasattr(core, 'currency'):
            currency = core.currency
        else:
            currency = "EUR"

        extension = Extension.construct()
        extension.url = "amount"
        money = Money(**{
            "value": imis_premium.amount,
            "currency": currency
        })
        extension.valueMoney = money
        asset_extensions.extension.append(extension)

    @classmethod
    def build_premium_receipt_ext(cls, asset_extensions, imis_premium):
        extension = Extension.construct()
        extension.url = "receipt"
        extension.valueString = imis_premium.receipt
        asset_extensions.extension.append(extension)

    @classmethod
    def build_premium_date_ext(cls, asset_extensions, imis_premium):
        extension = Extension.construct()
        extension.url = "date"
        extension.valueDate = imis_premium.pay_date
        asset_extensions.extension.append(extension)

    @classmethod
    def build_premium_type_ext(cls, asset_extensions, imis_premium):
        extension = Extension.construct()
        extension.url = "type"
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/contract-premium-type"
        extension.valueCodeableConcept = cls.build_codeable_concept(code=imis_premium.pay_type, system=system)
        if len(extension.valueCodeableConcept.coding) == 1:
            extension.valueCodeableConcept.coding[0].display = PayTypeMapping.pay_type[imis_premium.pay_type]
        asset_extensions.extension.append(extension)

    @classmethod
    def build_contract_asset_use_period(cls, contract_asset, imis_policy):
        period_use = Period.construct()
        period = Period.construct()
        if imis_policy.start_date is not None:
            period.start = imis_policy.start_date.strftime("%Y-%m-%d")
            period_use.start = period.start
        if imis_policy.effective_date is not None:
            period_use.start = imis_policy.effective_date.strftime("%Y-%m-%d")
            if period_use.start is None:
                period.start = period_use.start
        if imis_policy.expiry_date is not None:
            period_use.end = imis_policy.expiry_date.strftime("%Y-%m-%d")
            period.end = period_use.end

        if type(contract_asset.usePeriod) is not list:
            contract_asset.usePeriod = [period_use]
        else:
            contract_asset.usePeriod.append(period_use)
        if type(contract_asset.period) is not list:
            contract_asset.period = [period]
        else:
            contract_asset.period.append(period)
        return contract_asset

    @classmethod
    def build_contract_term_offer(cls, contract_term_offer, imis_policy, reference_type):
        offer = ContractTermOffer.construct()

        offer_party = ContractTermOfferParty.construct()
        offer_party.reference = [PatientConverter.build_fhir_resource_reference(imis_policy.family.head_insuree, 'Patient')]
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/contract-resource-party-role"
        offer_party.role = cls.build_codeable_concept(code="beneficiary", system=system)
        if len(offer_party.role.coding) == 1:
            offer_party.role.coding[0].display = _("Beneficiary")

        offer.party = [offer_party]
        contract_term_offer.offer = offer

    @classmethod
    def build_contract_status(cls, contract, imis_policy):
        if f"{imis_policy.status}" in ContractStatus.contract_status:
            contract.status = ContractStatus.contract_status[f"{imis_policy.status}"]
        else:
            contract.status = imis_policy.status
        return contract

    @classmethod
    def build_contract_state(cls, contract, imis_policy):
        if f"{imis_policy.stage}" in ContractState.contract_state:
            contract.legalState = cls.build_simple_codeable_concept(ContractState.contract_state[f"{imis_policy.stage}"])
        else:
            contract.legalState = cls.build_simple_codeable_concept(imis_policy.stage)
        return contract

    @classmethod
    def build_contract_valued_item_entity(cls, contract_asset, imis_policy):
        valued_item = ContractTermAssetValuedItem.construct()
        typeReference = cls.build_fhir_resource_reference(imis_policy.product, "InsurancePlan", imis_policy.product.code)
        valued_item.entityReference = typeReference
        policy_value = Money.construct()
        policy_value.value = imis_policy.value
        valued_item.net = policy_value
        if type(contract_asset.valuedItem) is not list:
            contract_asset.valuedItem = [valued_item]
        else:
            contract_asset.valuedItem.append(valued_item)
        return contract_asset

    @classmethod
    def build_contract_asset_type_reference(cls, contract_asset, imis_policy, reference_type):
        # type reference - take insurees covered as a policy patient
        from core import datetime
        now = datetime.datetime.now()

        list_insuree_policy = InsureePolicy.objects.filter(
            Q(policy=imis_policy),
            Q(validity_from__lte=now),
            Q(validity_to__isnull=True) | Q(validity_to__gte=now),
        ).only('insuree')

        for insuree_policy in list_insuree_policy:
            insuree = insuree_policy.insuree
            type_reference = cls.build_fhir_resource_reference(
                insuree, "Patient", insuree.chf_id, reference_type=reference_type
            )
            if type(contract_asset.typeReference) is not list:
                contract_asset.typeReference = [type_reference]
            else:
                contract_asset.typeReference.append(type_reference)

        return contract_asset

    @classmethod
    def build_imis_period(cls, imis_policy, fhir_contract, errors):
        for term in fhir_contract:
            if term.asset:
                for asset in term.asset:
                    if asset.period:
                        for period in asset.period:
                            if not cls.valid_condition(period.start is None, _('Missing  `period start` attribute'),errors):
                                imis_policy.start_date = TimeUtils.str_to_date(period.start)
                                imis_policy.enroll_date = TimeUtils.str_to_date(period.start)
                            if not cls.valid_condition(period.end is None, _('Missing  `period end` attribute'),errors):
                                imis_policy.expiry_date = TimeUtils.str_to_date(period.end)
                    else:
                        cls.valid_condition(not asset.period, _('Missing  `period` attribute'),errors)

    @classmethod
    def build_imis_useperiod(cls, imis_policy,fhir_contract,errors):
        for term in  fhir_contract:
            if term.asset:
                for asset in term.asset:
                    if asset.usePeriod:
                        for period in asset.usePeriod:
                            if not cls.valid_condition(period.start is None, _('Missing  `usePeriod start` attribute'),errors):
                                imis_policy.effective_date = TimeUtils.str_to_date(period.start)
                            if not cls.valid_condition(period.end is None, _('Missing  `usePeriod end` attribute'),errors):
                                imis_policy.expiry_date = TimeUtils.str_to_date(period.end)
                    else:
                        cls.valid_condition(not asset.usePeriod, _('Missing  `usePeriod` attribute'),errors)

    @classmethod
    def build_imis_status(cls, fhir_contract, imis_policy,errors):
        if fhir_contract.status:
            if fhir_contract.status == R4CoverageConfig.get_status_idle_code():
                imis_policy.status = ContractStatus.imis_map_status(R4CoverageConfig.get_status_idle_code(), imis_policy)
            elif fhir_contract.status == R4CoverageConfig.get_status_active_code():
                imis_policy.status = ContractStatus.imis_map_status(R4CoverageConfig.get_status_active_code(), imis_policy)
            elif fhir_contract.status == R4CoverageConfig.get_status_suspended_code():
                 imis_policy.status = ContractStatus.imis_map_status(R4CoverageConfig.get_status_suspended_code(), imis_policy)
            elif fhir_contract.status == R4CoverageConfig.get_status_expired_code():
                 imis_policy.status = ContractStatus.imis_map_status(R4CoverageConfig.get_status_expired_code(), imis_policy)
            else:
                imis_policy.status = ContractStatus.imis_map_status(R4CoverageConfig.get_status_idle_code(), imis_policy)
        else:
            cls.valid_condition(fhir_contract.status is None, _('Missing  `status` attribute'),errors)

    @classmethod
    def build_imis_author(cls, fhir_contract, imis_policy, errors):
        if fhir_contract.author:
            reference = fhir_contract.author.reference.split("Practitioner/", 2)
            imis_policy.officer = Officer.objects.get(uuid=reference[1])
        else:
            cls.valid_condition(not fhir_contract.author, _('Missing  `author` attribute'), errors)

    @classmethod
    def build_imis_subject(cls, fhir_contract, imis_policy, errors):
        if fhir_contract.subject:
            for subject in  fhir_contract.subject:
                if subject.reference is not None:
                    reference = subject.reference.split("Group/", 2)
                    try:
                        imis_policy.family = Family.objects.filter(uuid=reference[1]).first()
                    except:
                        cls.valid_condition(True, _('Missing  `Family head provided does not exist` attribute'), errors)
        else:
            cls.valid_condition(not fhir_contract.subject, _('Missing  `subject` attribute'), errors)

    @classmethod
    def build_imis_signer(cls, fhir_contract, imis_policy, errors):
        if fhir_contract.signer:
            for signer in  fhir_contract.signer:
                if signer.type:
                    if signer.type.text and signer.party.reference is not None:
                        if signer.type.text == 'HeadOfFamily':
                            reference = signer.party.reference.split("/", 2)
                            try:
                                insuree = Insuree.objects.get(uuid=reference[1])
                                if insuree.head:
                                    imis_policy.family= Family.objects.filter(head_insuree=insuree).first()
                                else:
                                    cls.valid_condition(True, _('Missing  `Member details provided belong to a depedant` attribute'),errors)
                            except:
                                cls.valid_condition(True, _('Missing  `Family head provided does not exist` attribute'),errors)
                        elif signer.type.text == 'EnrolmentOfficer':
                            reference = signer.party.reference.split("/", 2)
                            imis_policy.officer = Officer.objects.get(uuid=reference[1])
                        else:
                            pass
                else:
                    cls.valid_condition(signer.type is None, _('Missing  `type` attribute'),errors)
        else:
            cls.valid_condition(not fhir_contract.signer, _('Missing  `signer` attribute'),errors)

    @classmethod
    def build_imis_insurees(cls,fhir_contract,imis_policy,errors):
        if fhir_contract.term:
            insurees =[]
            for term in  fhir_contract.term:
                if term.asset:
                    for asset in term.asset:
                        if asset.typeReference:
                            for item in asset.typeReference:
                               if item.reference is not None:
                                   reference = item.reference.split("Patient/", 2)
                                   obj = Insuree.objects.get(uuid=reference[1])
                                   if imis_policy.family_id is not None:
                                       if obj.family == imis_policy.family:
                                           if type(insurees) is not list:
                                               insurees = [obj.uuid]
                                           else:
                                               insurees.append(obj.uuid)
                                       else:
                                            if 'Missing  `Invalid Context reference` attribute' not in errors:
                                                cls.valid_condition(True, _('Missing  `Invalid Context reference` attribute'),errors)
                            imis_policy.insurees = insurees
                        else:
                            cls.valid_condition(not asset.context, _('Missing  `context` attribute'),errors)
                else:
                    cls.valid_condition(not term.asset, _('Missing  `asset` attribute'),errors)

        else:
            cls.valid_condition(not fhir_contract, _('Missing  `term` attribute'),errors)

    @classmethod
    def build_imis_product(cls,fhir_contract, imis_policy, errors):
        if fhir_contract.term:
            for term in  fhir_contract.term:
                if term.asset:
                    for asset in term.asset:
                        if asset.valuedItem:
                            for item in asset.valuedItem:
                                if item.entityReference is not None:
                                    if item.entityReference.reference is not None:
                                        reference = item.entityReference.reference.split("InsurancePlan/", 2)
                                        imis_policy.product = Product.objects.get(uuid=reference[1])
                                if item.net is not None:
                                    if item.net.value is not None:
                                        imis_policy.value = item.net.value
                        else:
                            cls.valid_condition(not asset.valuedItem, _('Missing  `valuedItem` attribute'), errors)
                else:
                    cls.valid_condition(not term.asset, _('Missing  `asset` attribute'), errors)

        else:
            cls.valid_condition(not fhir_contract, _('Missing  `term` attribute'), errors)

    @classmethod
    def build_imis_state(cls,fhir_contract, imis_policy, errors):
        if fhir_contract.legalState:
            if fhir_contract.legalState.text:
                if fhir_contract.legalState.text == R4CoverageConfig.get_status_offered_code():
                    imis_policy.stage = ContractState.imis_map_stage(R4CoverageConfig.get_status_offered_code(), imis_policy)
                elif fhir_contract.legalState.text == R4CoverageConfig.get_status_renewed_code():
                    imis_policy.stage = ContractState.imis_map_stage(R4CoverageConfig.get_status_renewed_code(), imis_policy)
                else:
                    pass
        else:
            cls.valid_condition(fhir_contract.legalState is None, _('Missing  `legalState` attribute'), errors)

    @classmethod
    def build_imis_contributions(cls, fhir_contract, imis_policy, errors):
        premiums = []
        if fhir_contract.term:
            for term in fhir_contract.term:
                if term.asset:
                    for asset in term.asset:
                        if asset.extension and len(asset.extension) > 0:
                            if len(asset.extension[0].extension) > 0:
                                imis_contribution = Premium()
                                imis_contribution.uuid = None
                                contribution_extensions = asset.extension[0].extension
                                for fhir_contribution in contribution_extensions:
                                    cls.build_imis_contribution(fhir_contribution, imis_contribution)
                                premiums.append(imis_contribution)
        imis_policy.contributions = premiums

    @classmethod
    def build_imis_contribution(cls, fhir_contribution, imis_contribution):
        if fhir_contribution.url == "payer":
            cls.build_imis_contribution_payer(fhir_contribution, imis_contribution)
        if fhir_contribution.url == "amount":
            cls.build_imis_contribution_amount(fhir_contribution, imis_contribution)
        if fhir_contribution.url == "receipt":
            cls.build_imis_contribution_receipt(fhir_contribution, imis_contribution)
        if fhir_contribution.url == "date":
            cls.build_imis_contribution_pay_date(fhir_contribution, imis_contribution)
        if fhir_contribution.url == "type":
            cls.build_imis_contribution_pay_type(fhir_contribution, imis_contribution)

    @classmethod
    def build_imis_contribution_payer(cls, fhir_contribution, imis_contribution):
        # TODO add payer to contribution
        pass

    @classmethod
    def build_imis_contribution_amount(cls, fhir_contribution, imis_contribution):
        imis_contribution.amount = fhir_contribution.valueMoney.value

    @classmethod
    def build_imis_contribution_receipt(cls, fhir_contribution, imis_contribution):
        imis_contribution.receipt = fhir_contribution.valueString

    @classmethod
    def build_imis_contribution_pay_date(cls, fhir_contribution, imis_contribution):
        imis_contribution.pay_date = fhir_contribution.valueDate

    @classmethod
    def build_imis_contribution_pay_type(cls, fhir_contribution, imis_contribution):
        coding = fhir_contribution.valueCodeableConcept.coding
        if len(coding) > 0:
            code = coding[0]
            imis_contribution.pay_type = code.code



from unittest.mock import patch, MagicMock

from django.test import TestCase

from fhir.resources.fhirabstractmodel import FHIRAbstractModel

from api_fhir_r4.serializers import PatientSerializer


class ContainedContentHelper(object):
    @staticmethod
    def build_test_converter(returned_obj=FHIRAbstractModel.construct()):
        converter = MagicMock()
        converter.convert_from_source = MagicMock(name='convert_from_source', return_value=[returned_obj])
        return converter


class ContainedContentSerializerMixinTestCase(TestCase):
    from api_fhir_r4.containedResources.serializerMixin import ContainedContentSerializerMixin

    class BaseTestSerializer:
        context = {'contained': True}

        def to_representation(self, obj):
            return FHIRAbstractModel.construct().dict()

    class TestSerializer(ContainedContentSerializerMixin, BaseTestSerializer):

        @property
        def contained_resources(self):
            from api_fhir_r4.containedResources.containedResources import AbstractContainedResourceCollection
            class TestContainedResource(AbstractContainedResourceCollection):
                @classmethod
                def _definitions_for_serializers(cls):
                    from api_fhir_r4.containedResources.containedResources import ContainedResourceDefinition
                    return {
                        PatientSerializer: ContainedResourceDefinition('insuree', 'Patient')
                    }

            return TestContainedResource

    def test_resource_transformation(self):
        test_serializer = self.TestSerializer()
        test_imis_obj = MagicMock()
        test_imis_obj.insuree = MagicMock()
        representation = test_serializer.to_representation(test_imis_obj)

        expected_outcome = {'contained': []}
        self.assertEqual(dict(representation), expected_outcome)



from api_fhir_r4.configurations import GeneralConfiguration, R4LocationConfig
from api_fhir_r4.models.imisModelEnums import ImisLocationType


class LocationTypeMapping:
    SYSTEM = f'{GeneralConfiguration.get_system_base_url()}StructureDefinition/openimis-location'
    PHYSICAL_TYPE_SYSTEM = f'{GeneralConfiguration.get_system_base_url()}/CodeSystem/location-type'

    PHYSICAL_TYPES_DEFINITIONS = {
        ImisLocationType.REGION.value: {
            'code': R4LocationConfig.get_fhir_code_for_region(),
            'display': 'Region',
            'system': PHYSICAL_TYPE_SYSTEM
        },
        ImisLocationType.DISTRICT.value: {
            'code': R4LocationConfig.get_fhir_code_for_district(),
            'display': 'District',
            'system': PHYSICAL_TYPE_SYSTEM
        },
        ImisLocationType.WARD.value: {
            'code': R4LocationConfig.get_fhir_code_for_ward(),
            'display': 'Municipality/Ward',
            'system': PHYSICAL_TYPE_SYSTEM
        },
        ImisLocationType.VILLAGE.value: {
            'code': R4LocationConfig.get_fhir_code_for_village(),
            'display': 'Village',
            'system': PHYSICAL_TYPE_SYSTEM
        }
    }



import copy
from policyholder.models import PolicyHolder
from api_fhir_r4.converters import PolicyHolderOrganisationConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers import BaseFHIRSerializer


class PolicyHolderOrganisationSerializer(BaseFHIRSerializer):
    fhirConverter = PolicyHolderOrganisationConverter()

    def create(self, validated_data):
        if PolicyHolder.objects.filter(code=validated_data['code']).count() > 0:
            raise FHIRException('Exists Organization with following code `{}`'.format(validated_data['code']))
        validated_data.pop('_original_state')
        validated_data.pop('_state')
        request = self.context.get('request', None)
        if request:
            validated_data['user_created_id']=request.user.id
            validated_data['user_updated_id']=request.user.id
        obj=PolicyHolder(**validated_data)
        obj.save()
        return obj

    def update(self, instance, validated_data):
        request = self.context.get('request', None)
        if request:
            validated_data['user_updated_id'] = request.user.id
        instance.legal_form = validated_data.get('legal_form', instance.legal_form)
        instance.phone = validated_data.get('phone', instance.phone)
        instance.email = validated_data.get('email', instance.email)
        instance.fax = validated_data.get('fax', instance.fax)
        instance.trade_name = validated_data.get('trade_name', instance.trade_name)
        instance.address = validated_data.get('address', instance.address)
        instance.bank_account = validated_data.get('bank_account', instance.address)
        instance.accountancy_account = validated_data.get('accountancy_account', instance.accountancy_account)
        instance.user_updated_id = validated_data.get('user_updated_id', instance.user_updated_id)
        instance.save()
        return instance



import core
from dateutil import parser


class TimeUtils(object):

    @classmethod
    def now(cls):
        return core.datetime.datetime.now()

    @classmethod
    def date(cls):
        return core.datetime.datetime.date(cls.now())

    @classmethod
    def str_to_date(cls, str_date, str_time="00:00:00"):
        py_date = parser.parse(f"{str_date} {str_time}")
        return core.datetime.datetime.from_ad_datetime(py_date)

    @classmethod
    def str_iso_to_date(cls, str_iso_datetime):
        py_date = parser.parse(f"{str_iso_datetime}")
        return core.datetime.datetime.from_ad_datetime(py_date)



from functools import lru_cache

from claim.models import Claim
from core.models import User

from api_fhir_r4.converters import ClaimResponseConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


class ClaimResponseSerializer(BaseFHIRSerializer):

    fhirConverter = ClaimResponseConverter

    UPDATABLE_FIELDS = ['status', 'rejection_reason', 'date_from', 'date_to', 'feedback', 'visit_type',
                        'review_status', 'health_facility', 'adjustment', 'icd', 'icd_1', 'icd_2', 'icd_3', 'icd_4']

    def update(self, instance: Claim, validated_data):
        self._assign_values_if_available(instance, validated_data)
        self._save_claim_serviced_items(instance, validated_data)
        self._assign_audit_user_id(instance, self.get_audit_user_id())
        instance.save()
        return instance

    def _assign_values_if_available(self, instance, validated_data, fields=None):
        if fields is None:
            fields = self.UPDATABLE_FIELDS

        for next_field in fields:
            new_value = validated_data.get(next_field, None)
            if new_value:
                setattr(instance, next_field, new_value)

    def _save_claim_serviced_items(self, instance, validated_data):
        items = validated_data.get('claim_items', [])
        services = validated_data.get('claim_services', [])
        audit_user_id = self.get_audit_user_id()

        for serviced in items+services:
            serviced.claim = instance
            self._assign_audit_user_id(serviced, audit_user_id)
            serviced.save()

    def _assign_audit_user_id(self, model, audit_user_id):
        if isinstance(model.audit_user_id, int) and not isinstance(audit_user_id, int):
            # In case if model use Id instead of uuid
            audit_user = self._get_audit_user(audit_user_id)
            model.audit_user_id = audit_user.id
        else:
            model.audit_user_id = audit_user_id

    @lru_cache(maxsize=None)
    def _get_audit_user(self, audit_user_uuid: str) -> User:
        return User.objects.get(uuid=audit_user_uuid)



import json
import os

from core.models import User
from fhir.resources import construct_fhir_element
from rest_framework import status

from api_fhir_r4.configurations import R4IdentifierConfig
from api_fhir_r4.converters import BaseFHIRConverter
from fhir.resources.bundle import Bundle
from api_fhir_r4.utils import DbManagerUtils


class GenericFhirAPITestMixin(object):

    @property
    def base_url(self):
        return None

    @property
    def _test_json_path(self):
        return None

    _TEST_SUPERUSER_NAME = 'admin'
    _TEST_SUPERUSER_PASS = 'adminadmin'#'Admin123'
    _test_request_data = None

    def setUp(self):
        if self._test_json_path:
            dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
            json_representation = open(dir_path + self._test_json_path).read()
            self._test_request_data = json.loads(json_representation)

    def login(self):
        user = DbManagerUtils.get_object_or_none(User, username=self._TEST_SUPERUSER_NAME)
        if user is None:
            user = self.__create_superuser()
        self.client.force_authenticate(user=user)

    def __create_superuser(self):
        User.objects.create_superuser(username=self._TEST_SUPERUSER_NAME, password=self._TEST_SUPERUSER_PASS)
        return DbManagerUtils.get_object_or_none(User, username=self._TEST_SUPERUSER_NAME)

    def get_bundle_from_json_response(self, response):
        response_json = response.json()
        bundle = construct_fhir_element(response_json['resourceType'], response_json)
        self.assertTrue(isinstance(bundle, Bundle))
        return bundle

    def get_id_for_created_resource(self, response):
        result = None
        response_json = response.json()
        fhir_obj = construct_fhir_element(response_json['resourceType'], response_json)
        if hasattr(fhir_obj, 'identifier'):
            result = BaseFHIRConverter.get_fhir_identifier_by_code(fhir_obj.identifier,
                                                                   R4IdentifierConfig.get_fhir_uuid_type_code())
        return result

    def get_fhir_obj_from_json_response(self, response):
        response_json = response.json()
        fhir_obj = construct_fhir_element(response_json['resourceType'], response_json)
        return fhir_obj

    def test_get_should_required_login(self):
        response = self.client.get(self.base_url, data=None, format='json')
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)



"""
Basic building blocks for generic class based views.

We don't bind behaviour to http method handlers yet,
which allows mixin classes to be composed in interesting ways.
"""
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from itertools import chain
from typing import Dict, Type, Callable, Iterable, Tuple, List

from rest_framework import status
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.settings import api_settings
from rest_framework.serializers import Serializer
from rest_framework.exceptions import ValidationError, PermissionDenied
from django.http import Http404
from django.db.models.query import QuerySet
from django.core.exceptions import ObjectDoesNotExist, FieldError

from api_fhir_r4.permissions import FHIRApiPermissions

logger = logging.getLogger(__name__)


def _MultiserializerPermissionClassWrapper(PermissionClass):
    def has_permission(self, request, view, queryset):
        if getattr(view, '_ignore_model_permissions', False):
            return True

        if not request.user or (not request.user.is_authenticated and self.authenticated_users_only):
            return False

        perms = self.get_required_permissions(request.method, queryset.model)
        return request.user.has_perms(perms)

    permission_class = type('PermissionClassWrapper', PermissionClass.__bases__, dict(PermissionClass.__dict__))
    permission_class.has_permission = has_permission
    return permission_class


class GenericMultiSerializerViewsetMixin(ABC):

    @property
    def permission_classes(self):
        """
        Multi-serializer classes should use permissions added to registered serializers, instead for viewset directly.
        Initial permission check is used only to validate if request is validated with user.
        Returns:

        """
        return (IsAuthenticated,)

    def get_serializer_class(self):
        """
        serializer_class is not meant to be used in Multiserializer viewset context
        """
        pass

    @property
    def serializer_class(self):
        raise NotImplementedError("serializer_class is not meant to be used in Multiserializer viewset context")

    @property
    @abstractmethod
    def serializers(self) \
            -> Dict[Type[Serializer], Tuple[Callable[[], QuerySet], Callable[[Dict], bool], Tuple[FHIRApiPermissions]]]:
        """
        Variable used for determining serializers available for the given viewset. It's a dictionary where keys
        are serializers and values are tuples with two functions.
        First one is responsible for returning queryset used by the serializer.
        Second one is validator function responsible for determining if given serializer
        is eligible for request context.
        Third element is tuple of permissions, as ViewSet doesn't provide permission check for endpoint it's required
        to validate permissions against specific serializers. If user doesn't have permission for any serializer then
        401 is raised.
        Returns:

        """
        raise NotImplementedError('serializers method has to return dictionary of serializers')

    def get_eligible_serializers(self) -> List[Type[Serializer]]:
        eligible = []
        context = self.get_serializer_context()

        eligible_from_permissions = self._get_eligible_from_user_permissions()

        for serializer, (queryset, eligibility_validator, permission_class) in self.serializers.items():
            if eligibility_validator(context) and serializer in eligible_from_permissions:
                eligible.append(serializer)
        return eligible

    def _aggregate_results(self, results):
        """
        It's expected for serializers to aggregate output data in format that will be accepted by
        rest_framework.Response as data argument.
        By default it returns first argument
        :param results: Results for execution of actions on individual serializers
        :return: By default returns result in raw format.
        """
        return results

    def get_object_by_queryset(self, qs):
        """
        Adjusted Django Rest Framework's GenericAPIView
        """

        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field

        assert lookup_url_kwarg in self.kwargs, (
            'Expected view %s to be called with a URL keyword argument '
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            'attribute on the view correctly.' %
            (self.__class__.__name__, lookup_url_kwarg)
        )

        queryset = self.filter_queryset(qs)

        try:
            filter_kwargs = {self.lookup_field: self.kwargs[lookup_url_kwarg]}
            obj = queryset.get(**filter_kwargs)
            # May raise a permission denied
            self.check_object_permissions(self.request, obj)
        except (ObjectDoesNotExist, PermissionDenied):
            return None
        except FieldError as e:
            logger.warning(F"Fetching object with multiserializer has failed, lookup field {self.lookup_field} does"
                           F"not exist for {queryset.model}: {str(e)}")
            return None
        return obj

    def validate_single_eligible_serializer(self):
        eligible_serializers = len(self.get_eligible_serializers())
        if eligible_serializers == 0:
            self._raise_no_eligible_serializer()
        if eligible_serializers > 1:
            self._raise_multiple_eligible_serializers()

    def get_eligible_serializers_iterator(self):
        for serializer in self.get_eligible_serializers():
            yield serializer, self.serializers[serializer]

    def _raise_no_eligible_serializer(self):
        raise AssertionError("Failed to match serializer eligible for given request")

    def _raise_multiple_eligible_serializers(self):
        raise AssertionError("Ambiguous request, more than one serializer is eligible for given action")

    def _get_eligible_from_user_permissions(self):
        eligible_serializers = []
        for serializer, (queryset, eligibility_validator, permission_classes) in self.serializers.items():
            permission_classes = [
                _MultiserializerPermissionClassWrapper(perm_cls)() for perm_cls in permission_classes
            ]
            if all([p.has_permission(self.request, self, queryset) for p in permission_classes]):
                eligible_serializers.append(serializer)

        if len(eligible_serializers) == 0:
            self.permission_denied(
                self.request,
                message="User unauthorized for any of the resourceType available in the view."
            )

        return eligible_serializers


class MultiSerializerCreateModelMixin(GenericMultiSerializerViewsetMixin, ABC):
    """
    Create a model instance.
    """
    def create(self, request, *args, **kwargs):
        self._validate_create_request()
        results = []
        for serializer, _ in self.get_eligible_serializers_iterator():
            data = self._create_for_serializer(serializer, request, *args, **kwargs)
            results.append(data)

        headers = self.get_success_headers(results)
        response = results[0]  # By default there should be only one eligible serializer
        return Response(response, status=status.HTTP_201_CREATED, headers=headers)

    def _create_for_serializer(self, serializer, request, *args, **kwargs):
        context = self.get_serializer_context()  # Required for audit user id
        serializer = serializer(data=request.data, context=context)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        return serializer.data

    def perform_create(self, serializer):
        serializer.save()

    def get_success_headers(self, res):
        try:
            return {'Location': str([data[api_settings.URL_FIELD_NAME] for data in res])}
        except (TypeError, KeyError):
            return {}

    def _validate_create_request(self):
        self.validate_single_eligible_serializer()


class _JoinedQuerysets:
    def __init__(self, *qs):
        self.querysets = qs

    def __iter__(self):
        return chain(self.querysets)

    def __len__(self):
        return sum([len(qs) for qs in self.querysets])

    def __getitem__(self, k):
        if not isinstance(k, (int, slice)):
            raise TypeError

        return self.__get_queryset_for_item(k)

    def __get_queryset_for_item(self, k):
        if isinstance(k, int):
            return self.__get_queryset_for_index(k)
        else:
            # slice
            if k.step:
                raise ValidationError("Step not supported in joined queryset context.")
            start, end = int(k.start) if k else None, int(k.stop) if k else None
            intersection = [None, None]
            final_query = []
            if start:
                start_qs, qs_idx = self.__get_queryset_for_index(start)
                intersection[0] = self.querysets.index(start_qs) + 1
                start_qs = start_qs.all()[:qs_idx]
            if end:
                end_qs, qs_idx = self.__get_queryset_for_index(end)
                intersection[1] = self.querysets.index(end_qs)
                end_qs = end_qs.all()[:qs_idx]

            if start:
                final_query.append(start_qs)

            final_query.extend(self.querysets[intersection[0]:intersection[1]])

            if end:
                final_query.append(end_qs)

            return list(chain(*final_query))

    def __get_queryset_for_index(self, k):
        """
        Return queryset for which given index is relevant. If given index is out of range it raises IndexEr
        Args:
            k: index of element.

        Returns:
            Tuple of queryset and index k relative for given queryset

        """
        next_queryset_last_index = 0
        for qs in self.querysets:
            qs_len = qs.count()
            next_queryset_last_index += qs_len
            if k < next_queryset_last_index:
                index_in_queryset = k - (next_queryset_last_index - qs_len)
                return qs, index_in_queryset
        raise IndexError

    def count(self):
        return sum([qs.count() for qs in self.querysets])


class MultiSerializerListModelMixin(GenericMultiSerializerViewsetMixin, ABC):
    """
    List a queryset.
    """
    def list(self, request, *args, **kwargs):
        self._validate_list_model_request()
        filtered_querysets = {}  # {serialzer: qs}

        for serializer, (qs, _, _) in self.get_eligible_serializers_iterator():
            next_serializer_data = self.filter_queryset(qs)
            model = next_serializer_data.model
            filtered_querysets[model, serializer] = next_serializer_data

        if not filtered_querysets:
            return {}

        try:
            querysets = self._join_querysets([*filtered_querysets.values()])
            page = self.paginate_queryset(querysets)
            data = self.__dispatch_page_data(page)
            serialized_data = self._serialize_dispatched_data(data, dict(filtered_querysets.keys()))
            data = self.get_paginated_response(serialized_data)
            return data
        except Exception as e:
            import traceback
            traceback.print_exc()
            raise e

    def _validate_list_model_request(self):
        # By default always valid
        return True

    def __dispatch_page_data(self, page):
        x = defaultdict(list)
        for r in page:
            x[type(r)].append(r)
        return x

    def _serialize_dispatched_data(self, data, serializer_models):
        serialized = []
        for model, model_data in data.items():
            serializer_cls = serializer_models.get(model, None)
            if not serializer_cls:
                logger.error(f"Found data of type {model_data} but it couldn't be matched with "
                             f"any of available serializers {serializer_models}")
                continue
            else:
                serializer = serializer_cls(tuple(model_data), many=True)
                serialized.extend(serializer.data)

        return serialized

    def _join_querysets(self, querysets: List[QuerySet]):
        if len(querysets) == 0:
            raise ValueError("At least one eligible queryset required.")
        elif len(querysets) == 1:
            return querysets[0]
        else:
            # Inefficient due to pulling every entry and not only paginated chunk
            chained = _JoinedQuerysets(*querysets)
            # Chain doesn't provide len, but it's required by paginator
            return chained


class MultiSerializerRetrieveModelMixin(GenericMultiSerializerViewsetMixin, ABC):
    """
    Retrieve a model instance.
    """
    def retrieve(self, request, *args, **kwargs):
        self._validate_retrieve_model_request()
        retrieved = []
        for serializer, (qs, _, _) in self.get_eligible_serializers_iterator():
            instance = self.get_object_by_queryset(qs=qs)
            serializer = serializer(instance)
            if serializer.data:
                retrieved.append(serializer.data)

        if len(retrieved) > 1:
            raise ValueError("Ambiguous retrieve result, object found for multiple serializers.")
        if len(retrieved) == 0:
            raise Http404

        return Response(retrieved[0])

    def _validate_retrieve_model_request(self):
        # By default always valid
        return True


class MultiSerializerUpdateModelMixin(GenericMultiSerializerViewsetMixin, ABC):
    """
    Update a model instance.
    """
    def update(self, request, *args, **kwargs):
        self._validate_update_request()
        partial = kwargs.pop('partial', False)
        results = []
        for serializer, (qs, _, _) in self.get_eligible_serializers_iterator():
            instance = self.get_object_by_queryset(qs=qs)
            update_result = self._update_for_serializer(serializer, instance, request.data, partial)
            results.append(update_result)

        response = results[0]  # By default there should be only one eligible serializer
        return Response(response)

    def _update_for_serializer(self, serializer, instance, data, partial, *args, **kwargs):
        context = self.get_serializer_context()  # Required for audit user id
        serializer = serializer(instance, data=data, partial=partial, context=context, *args, **kwargs)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)
        if getattr(instance, '_prefetched_objects_cache', None):
            instance._prefetched_objects_cache = {}
        return serializer.data

    def perform_update(self, serializer):
        serializer.save()

    def partial_update(self, request, *args, **kwargs):
        kwargs['partial'] = True
        return self.update(request, *args, **kwargs)

    def _validate_update_request(self):
        self.validate_single_updatable_resource()

    def validate_single_updatable_resource(self):
        instance = None
        for serializer, (qs, _, _) in self.get_eligible_serializers_iterator():
            obj = self.get_object_by_queryset(qs=qs)
            if obj and instance:
                # If more than one updatable instance found
                return False
            elif obj:
                instance = obj
        return True



import json
import os

from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from api_fhir_r4.utils import DbManagerUtils
from rest_framework.test import APITestCase
from rest_framework import status
from core.models import User
from core.services import create_or_update_interactive_user, create_or_update_core_user
from api_fhir_r4.tests import GenericFhirAPITestMixin
from api_fhir_r4.configurations import GeneralConfiguration
from core.test_helpers import create_test_officer
from insuree.test_helpers import create_test_insuree
from product.test_helpers import create_test_product


class ContractAPITests(GenericFhirAPITestMixin, APITestCase, LogInMixin):

    base_url = GeneralConfiguration.get_base_url()+'Contract/'
    _test_json_path = "/test/test_contract.json"

    _TEST_FAMILY_UUID = "e8bbb7e4-19ef-4bef-9342-9ab6b9a928d3"
    _TEST_OFFICER_UUID = "ff7db42d-874b-400a-bba7-e59b273ae123"
    _TEST_INSUREE_UUID = "f8c56ada-d76d-4f6c-aad3-cfddc9fb38eb"
    _TEST_PRODUCT_CODE = "TE123"
    _TEST_PRODUCT_UUID = "8ed8d2d9-2644-4d29-ba37-ab772386cfca"

    _test_json_path_credentials = "/tests/test/test_login.json"
    _test_request_data_credentials = None

    def setUp(self):
        super(ContractAPITests, self).setUp()
        dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        json_representation = open(dir_path + self._test_json_path_credentials).read()
        self._test_request_data_credentials = json.loads(json_representation)
        self.get_or_create_user_api()

    def create_dependencies(self):
        # create mocked insuree
        imis_insuree = create_test_insuree(with_family=True)
        imis_insuree.uuid = self._TEST_INSUREE_UUID
        imis_insuree.save()

        # update family uuid
        imis_family = imis_insuree.family
        imis_family.uuid = self._TEST_FAMILY_UUID
        imis_family.save()

        # create mocked product
        imis_product = create_test_product(self._TEST_PRODUCT_CODE, valid=True, custom_props=None)
        imis_product.uuid = self._TEST_PRODUCT_UUID
        imis_product.save()

        # create mocked officer
        imis_officer = create_test_officer()
        imis_officer.uuid = self._TEST_OFFICER_UUID
        imis_officer.save()

    def test_post_should_create_correctly(self):
        self.create_dependencies()
        response = self.client.post(
            GeneralConfiguration.get_base_url() + 'login/', data=self._test_request_data_credentials, format='json'
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}"
        }
        response = self.client.post(self.base_url, data=self._test_request_data, format='json', **headers)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertIsNotNone(response.content)



from claim.models import ClaimAdmin
from django.utils.translation import gettext as _

from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import ClaimAdminPractitionerConverter
from fhir.resources.contactpoint import ContactPoint
from fhir.resources.humanname import HumanName
from fhir.resources.identifier import Identifier
from fhir.resources.practitioner import Practitioner, PractitionerQualification
from api_fhir_r4.models.imisModelEnums import ContactPointSystem, ContactPointUse
from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.utils import TimeUtils


class ClaimAdminPractitionerTestMixin(GenericTestMixin):

    _TEST_LAST_NAME = "Smith"
    _TEST_OTHER_NAME = "John"
    _TEST_DOB = "1990-03-24"
    _TEST_ID = 1
    _TEST_UUID = "254f6268-964b-4d8d-aa26-20081f22235e"
    _TEST_CODE = "1234abcd"
    _TEST_PHONE = "813-996-476"
    _TEST_EMAIL = "TEST@TEST.com"

    def create_test_imis_instance(self):
        imis_claim_admin = ClaimAdmin()
        imis_claim_admin.last_name = self._TEST_LAST_NAME
        imis_claim_admin.other_names = self._TEST_OTHER_NAME
        imis_claim_admin.id = self._TEST_ID
        imis_claim_admin.uuid = self._TEST_UUID
        imis_claim_admin.code = self._TEST_CODE
        imis_claim_admin.dob = TimeUtils.str_to_date(self._TEST_DOB)
        imis_claim_admin.phone = self._TEST_PHONE
        imis_claim_admin.email_id = self._TEST_EMAIL
        return imis_claim_admin

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_LAST_NAME, imis_obj.last_name)
        self.assertEqual(self._TEST_OTHER_NAME, imis_obj.other_names)
        self.assertEqual(self._TEST_CODE, imis_obj.code)
        self.assertEqual(self._TEST_DOB+"T00:00:00", imis_obj.dob.isoformat())
        self.assertEqual(self._TEST_PHONE, imis_obj.phone)
        self.assertEqual(self._TEST_EMAIL, imis_obj.email_id)

    def create_test_fhir_instance(self):
        fhir_practitioner = Practitioner.construct()
        name = HumanName.construct()
        name.family = self._TEST_LAST_NAME
        name.given = [self._TEST_OTHER_NAME]
        name.use = "usual"
        fhir_practitioner.name = [name]
        identifiers = []
        code = ClaimAdminPractitionerConverter.build_fhir_identifier(
            self._TEST_CODE,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code()
        )
        identifiers.append(code)
        fhir_practitioner.identifier = identifiers
        fhir_practitioner.birthDate = self._TEST_DOB
        telecom = []
        phone = ClaimAdminPractitionerConverter.build_fhir_contact_point(
            self._TEST_PHONE,
            ContactPointSystem.PHONE,
            ContactPointUse.HOME
        )
        telecom.append(phone)
        email = ClaimAdminPractitionerConverter.build_fhir_contact_point(
            self._TEST_EMAIL,
            ContactPointSystem.EMAIL,
            ContactPointUse.HOME
        )
        telecom.append(email)
        fhir_practitioner.telecom = telecom

        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/practitioner-qualification-type"
        qualification = PractitionerQualification.construct()
        qualification.code = ClaimAdminPractitionerConverter.build_codeable_concept(
            system=system,
            code="CA",
            display=_("Claim Administrator")
        )
        fhir_practitioner.qualification = [qualification]

        return fhir_practitioner

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(1, len(fhir_obj.name))
        human_name = fhir_obj.name[0]
        self.assertTrue(isinstance(human_name, HumanName))
        self.assertEqual(self._TEST_OTHER_NAME, human_name.given[0])
        self.assertEqual(self._TEST_LAST_NAME, human_name.family)
        self.assertEqual("usual", human_name.use)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = ClaimAdminPractitionerConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_generic_type_code():
                self.assertEqual(self._TEST_CODE, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(self._TEST_UUID, identifier.value)
        self.assertEqual(self._TEST_DOB, fhir_obj.birthDate.isoformat())
        self.assertEqual(2, len(fhir_obj.telecom))
        for telecom in fhir_obj.telecom:
            self.assertTrue(isinstance(telecom, ContactPoint))
            if telecom.system == "phone":
                self.assertEqual(self._TEST_PHONE, telecom.value)
            elif telecom.system == "email":
                self.assertEqual(self._TEST_EMAIL, telecom.value)
        self.assertEqual(1, len(fhir_obj.qualification))
        self.assertEqual("CA", fhir_obj.qualification[0].code.coding[0].code)
        self.assertEqual("Claim Administrator", fhir_obj.qualification[0].code.coding[0].display)



import asyncio
import datetime

from asynctest import CoroutineMock, MagicMock, patch, call
from asgiref.sync import async_to_sync
from django.test import TestCase

from api_fhir_r4.models import Subscription
from api_fhir_r4.subscriptions.notificationClient import RestSubscriptionNotificationClient, \
    SubscriberNotificationOutput
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from aiohttp import web


class AsyncContextManagerMock(MagicMock):

    async def __aenter__(self):
        return self.aenter

    async def __aexit__(self, *args):
        pass


class TestSubscriptionNotificationClient(LogInMixin, TestCase):
    TEST_HEADERS_1 = """{"test-header": "123", "Authentication": "Bearer ABF13816"}"""
    EXPECTED_HEADER_1 = {'content-type': 'application/json', 'accept': 'application/json', 'test-header': '123', 'Authentication': 'Bearer ABF13816'}
    TEST_HEADERS_2 = """{"test-header": "123", "Authentication": "Bearer 61831FAB"}"""
    EXPECTED_HEADER_2 = {'content-type': 'application/json', 'accept': 'application/json', 'test-header': '123', 'Authentication': 'Bearer 61831FAB'}
    NOTIFICATION_CONTENT = {'notification_content': 'content'}

    def setUp(self) -> None:
        super().setUp()
        self._test_user = self.get_or_create_user_api()
        self._test_subscriptions = self._create_test_subscriptions()

    @async_to_sync
    @patch("api_fhir_r4.subscriptions.notificationClient.aiohttp.ClientSession.post")
    async def test_post_should_propagate_correctly(self, session):
        session.return_value.__aenter__.return_value.json = CoroutineMock(side_effect=
            [{'Notification': 'Thanks for notification'}, {'Notification': 'Thanks for notification'}])
        session.return_value.__aenter__.return_value.status = 200
        sub_client = RestSubscriptionNotificationClient()
        response = await sub_client.propagate_notifications_async(self.NOTIFICATION_CONTENT, self._test_subscriptions)

        expected = [SubscriberNotificationOutput(self._test_subscriptions[0], True, None),
                    SubscriberNotificationOutput(self._test_subscriptions[1], True, None)]
        self.assertListEqual(expected, list(response))

        session.assert_any_call(
            url='http://test-subscription-endpoint.io/post_uri/',
            headers=self.EXPECTED_HEADER_1, data=b'{"notification_content":"content"}')
        session.assert_any_call(
            url='http://test-subscription-endpoint.io/post_uri/',
            headers=self.EXPECTED_HEADER_2, data=b'{"notification_content":"content"}')

    @async_to_sync
    @patch("api_fhir_r4.subscriptions.notificationClient.aiohttp.ClientSession.post")
    async def test_post_server_unavailable(self, session):
        server_response = {'Notification': 'Server offline'}
        session.return_value.__aenter__.return_value.json = \
            CoroutineMock(side_effect=[server_response, server_response])
        session.return_value.__aenter__.return_value.status = 503

        sub_client = RestSubscriptionNotificationClient()
        response = await sub_client.propagate_notifications_async(self.NOTIFICATION_CONTENT, self._test_subscriptions)
        expected = [SubscriberNotificationOutput(self._test_subscriptions[0], False, server_response),
                    SubscriberNotificationOutput(self._test_subscriptions[1], False, server_response)]

        self.assertListEqual(expected, list(response))

        session.assert_any_call(
            url='http://test-subscription-endpoint.io/post_uri/',
            headers=self.EXPECTED_HEADER_1, data=b'{"notification_content":"content"}')
        session.assert_any_call(
            url='http://test-subscription-endpoint.io/post_uri/',
            headers=self.EXPECTED_HEADER_2, data=b'{"notification_content":"content"}')

    def _create_test_subscriptions(self):
        _valid_subscription = [self._create_valid(self.TEST_HEADERS_1), self._create_valid(self.TEST_HEADERS_2)]
        return _valid_subscription

    def _create_valid(self, headers):
        sub = Subscription(
            status=1, channel=0, endpoint='http://test-subscription-endpoint.io/post_uri/',
            headers=headers, expiring=datetime.datetime.now() + datetime.timedelta(days=10)
        )
        sub.save(username=self._test_user.username)
        return sub

    def _mock_coro(self):
        f = asyncio.Future()
        f.set_result(web.Response(text='Thanks for notification.'))
        return f



from api_fhir_r4.configurations import BaseApiFhirConfiguration, R4IdentifierConfig, \
    R4LocationConfig, R4MaritalConfig, R4IssueTypeConfig, R4ClaimConfig, R4CoverageEligibilityConfiguration, \
    R4CommunicationRequestConfig, R4OrganisationConfig, R4CoverageConfig, R4SubscriptionConfig


class R4ApiFhirConfig(BaseApiFhirConfiguration):

    @classmethod
    def get_identifier_configuration(cls):
        return R4IdentifierConfig

    @classmethod
    def get_location_type_configuration(cls):
        return R4LocationConfig

    @classmethod
    def get_marital_type_configuration(cls):
        return R4MaritalConfig

    @classmethod
    def get_issue_type_configuration(cls):
        return R4IssueTypeConfig

    @classmethod
    def get_claim_configuration(cls):
        return R4ClaimConfig

    @classmethod
    def get_coverage_eligibility_configuration(cls):
        return R4CoverageEligibilityConfiguration

    @classmethod
    def get_communication_request_configuration(cls):
        return R4CommunicationRequestConfig

    @classmethod
    def get_coverage_configuration(cls):
        return R4CoverageConfig

    @classmethod
    def get_organisation_configuration(cls):
        return R4OrganisationConfig

    @classmethod
    def get_subscription_configuration(cls):
        return R4SubscriptionConfig



import copy

from insuree.models import Family, Insuree
from api_fhir_r4.converters import GroupConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers import BaseFHIRSerializer
from insuree.services import FamilyService, InsureeService


class GroupSerializer(BaseFHIRSerializer):
    fhirConverter = GroupConverter()

    def create(self, validated_data):
        request = self.context.get("request")
        user = request.user

        insuree_id = validated_data.get('head_insuree_id')
        members_family = validated_data.pop('members_family')

        if Family.objects.filter(head_insuree_id=insuree_id).count() > 0:
            raise FHIRException('Exists family with the provided head')

        insuree = Insuree.objects.get(id=insuree_id)
        copied_data = copy.deepcopy(validated_data)
        copied_data["head_insuree"] = insuree.__dict__
        copied_data["contribution"] = None
        del copied_data['_state']

        new_family = FamilyService(user).create_or_update(copied_data)

        # assign members of family (insuree) to the family
        for mf in members_family:
            mf = mf.__dict__
            del mf['_state']
            mf['family_id'] = new_family.id
            InsureeService(user).create_or_update(mf)

        return new_family

    def update(self, instance, validated_data):
        # TODO: This doesn't work
        request = self.context.get("request")
        user = request.user
        chf_id = validated_data.get('chf_id')
        if Family.objects.filter(head_insuree_id=chf_id).count() == 0:
            raise FHIRException('No family with following chfid `{}`'.format(chf_id))
        family = Family.objects.get(head_insuree_id=chf_id, validity_to__isnull=True)
        validated_data["id"] = family.id
        validated_data["uuid"] = family.uuid
        del validated_data['_state']
        instance = FamilyService(user).create_or_update(validated_data)
        return instance



# TODO uncomment if someone need this serializer with connection to openHIM

"""
import logging
import os
import requests
from rest_framework.response import Response
from requests.auth import HTTPBasicAuth
from rest_framework import status
from policy.services import ByInsureeRequest, ByInsureeService, ByInsureeResponse
from api_fhir_r4.converters import PolicyCoverageEligibilityRequestConverter
from api_fhir_r4.serializers import BaseFHIRSerializer
"""

"""
class PolicyCoverageEligibilityRequestSerializer(BaseFHIRSerializer):

    fhirConverter = PolicyCoverageEligibilityRequestConverter()
    logger = logging.getLogger(__name__)
    def create(self, validated_data):
        data = self.context.get("request").data
        data['resourceType'] = "CoverageEligibilityRequest"
        url=os.environ.get('OPEHNHIM_URL')
        technical_user =os.environ.get('OPEHNHIM_USER')
        password = os.environ.get('OPEHNHIM_PASSWORD')
        try:
            response = requests.post(url+'CoverageEligibilityRequest',json=data,auth=HTTPBasicAuth(technical_user,password))
            if response.status_code == 200 and len(response.json())>1:
                res = response.json()
                return {"items":res['insurance']['item']}
            else:
                return self.eligibility_check(validated_data.get('chf_id'))
        except:
            return self.eligibility_check(validated_data.get('chf_id'))
        
    def eligibility_check(self,data):
        eligibility_request = ByInsureeRequest(chf_id=data)
        request = self.context.get("request")
        try:
            response = ByInsureeService(request.user).request(eligibility_request)            
        except TypeError:
            self.logger.warning('The insuree with chfid `{}` is not connected with policy. '
                                'The default eligibility response will be used.'
                                .format(validated_data.get('chfid')))
            response = self.create_default_eligibility_response()
        return response

    def create_default_eligibility_response(self):
        return ByInsureeResponse(
            by_insuree_request=None,
            items=[]
        )
    
    def create_eligibility_response(self, by_insuree_request,items):
        return ByInsureeResponse(
            by_insuree_request= by_insuree_request,
            items= items
        )
"""



from api_fhir_r4.converters import BillInvoiceConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


class BillSerializer(BaseFHIRSerializer):
    fhirConverter = BillInvoiceConverter()

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass



import uuid
from abc import ABC, abstractmethod, abstractproperty
from typing import Union

from django.db.models.query import QuerySet
from django.db.models import Model

from api_fhir_r4.converters import ReferenceConverterMixin


class GenericModelRetriever(ABC):

    @property
    @abstractmethod
    def identifier_field(self) -> str:
        pass

    @property
    @abstractmethod
    def serializer_reference_type(self) -> Union[
        ReferenceConverterMixin.UUID_REFERENCE_TYPE,
        ReferenceConverterMixin.CODE_REFERENCE_TYPE,
        ReferenceConverterMixin.DB_ID_REFERENCE_TYPE
    ]:
        pass

    @classmethod
    @abstractmethod
    def identifier_validator(cls, identifier_value) -> bool:
        pass

    @classmethod
    def retriever_additional_queryset_filtering(cls, queryset):
        # By default no additional changes are made in queryset
        return queryset

    @classmethod
    def get_model_object(cls, queryset: QuerySet, identifier_value) -> Model:
        return queryset.get(**{cls.identifier_field: identifier_value})


class UUIDIdentifierModelRetriever(GenericModelRetriever):
    identifier_field = 'uuid'
    serializer_reference_type = ReferenceConverterMixin.UUID_REFERENCE_TYPE

    @classmethod
    def identifier_validator(cls, identifier_value):
        return cls._is_uuid_identifier(identifier_value)

    @classmethod
    def _is_uuid_identifier(cls, identifier):
        try:
            uuid.UUID(str(identifier))
            return True
        except ValueError:
            return False


class DatabaseIdentifierModelRetriever(GenericModelRetriever):
    identifier_field = 'id'
    serializer_reference_type = ReferenceConverterMixin.DB_ID_REFERENCE_TYPE

    @classmethod
    def identifier_validator(cls, identifier_value):
        return isinstance(identifier_value, int) or identifier_value.isdigit()


class CodeIdentifierModelRetriever(GenericModelRetriever):
    identifier_field = 'code'
    serializer_reference_type = ReferenceConverterMixin.CODE_REFERENCE_TYPE

    @classmethod
    def identifier_validator(cls, identifier_value):
        return isinstance(identifier_value, str)

    @classmethod
    def add_retriever_queryset_filtering(cls, queryset):
        # By default no additional changes are made in queryset
        return queryset.filter(validity_to__is_null=True)


class CHFIdentifierModelRetriever(CodeIdentifierModelRetriever):
    identifier_field = 'chf_id'

    @classmethod
    def identifier_validator(cls, identifier_value):
        # From model specification
        return isinstance(identifier_value, str) and len(identifier_value) <= 12

    @classmethod
    def get_model_object(cls, queryset: QuerySet, identifier_value) -> Model:
        return queryset.get(**{cls.identifier_field: identifier_value, 'validity_to__isnull': True})


class GroupIdentifierModelRetriever(CHFIdentifierModelRetriever):
    identifier_field = 'head_insuree_id__chf_id'





from collections import OrderedDict
from django.urls import include, path
from drf_spectacular.views import SpectacularAPIView, SpectacularSwaggerView, SpectacularRedocView
from rest_framework.routers import DefaultRouter
from openIMIS.openimisapps import openimis_apps


from api_fhir_r4.views import LoginView, fhir as fhir_viewsets

imis_modules = openimis_apps()

router = DefaultRouter()
router.register(r'login', LoginView, basename="login")
router.register(r'Subscription', fhir_viewsets.SubscriptionViewSet, basename='Subscription_R4')

# register endpoint related to Product module if used
if 'product' in imis_modules:
    router.register(r'InsurancePlan', fhir_viewsets.ProductViewSet, basename="InsurancePlan_R4")

# register endpoint related to Location module if used
if 'location' in imis_modules:
    router.register(r'Location', fhir_viewsets.LocationViewSet, basename="Location_R4")
    # code system for openimis organization legal form
    router.register(
        r'CodeSystem/organization-hf-legal-form',
        fhir_viewsets.CodeSystemOrganizationHFLegalFormViewSet,
        basename="CodeSystem/organization-hf-legal-form_R4"
    )
    router.register(
        r'CodeSystem/organization-hf-level',
        fhir_viewsets.CodeSystemOrganizationHFLevelViewSet,
        basename="CodeSystem/organization-hf-level_R4"
    )

# register endpoint for insuree if used
if 'insuree' in imis_modules:
    router.register(r'Patient', fhir_viewsets.InsureeViewSet, basename="Patient_R4")
    router.register(r'Group', fhir_viewsets.GroupViewSet, basename="Group_R4")
    router.register(
        r'CoverageEligibilityRequest', fhir_viewsets.CoverageEligibilityRequestViewSet,
        basename="CoverageEligibilityRequest_R4"
    )
    # code system for openimis patient
    router.register(
        r'CodeSystem/patient-education-level',
        fhir_viewsets.CodeSystemOpenIMISPatientEducationLevelViewSet,
        basename="CodeSystem/patient-education-level_R4"
    )
    router.register(
        r'CodeSystem/patient-profession',
        fhir_viewsets.CodeSystemOpenIMISPatientProfessionViewSet,
        basename="CodeSystem/patient-profession_R4"
    )
    router.register(
        r'CodeSystem/patient-identification-type',
        fhir_viewsets.CodeSystemOpenIMISPatientIdentificationTypeViewSet,
        basename="CodeSystem/patient-identification-type_R4"
    )
    router.register(
        r'CodeSystem/patient-contact-relationship',
        fhir_viewsets.CodeSystemOpenIMISPatientRelationshipViewSet,
        basename="CodeSystem/patient-contact-relationship_R4"
    )
    # code system for openimis group
    router.register(
        r'CodeSystem/group-type',
        fhir_viewsets.CodeSystemOpenIMISGroupTypeViewSet,
        basename="CodeSystem/group-type_R4"
    )
    router.register(
        r'CodeSystem/group-confirmation-type',
        fhir_viewsets.CodeSystemOpenIMISGroupConfirmationTypeViewSet,
        basename="CodeSystem/group-confirmation-type_R4"
    )

# register endpoints related to medical module
if 'medical' in imis_modules:
    router.register(r'Medication', fhir_viewsets.MedicationViewSet, basename="Medication_R4")
    router.register(r'ActivityDefinition', fhir_viewsets.ActivityDefinitionViewSet, basename="ActivityDefinition_R4")

# register all endpoints related to c based on c
if 'claim' in imis_modules:
    router.register(r'Claim', fhir_viewsets.ClaimViewSet, basename="Claim_R4")
    router.register(r'ClaimResponse', fhir_viewsets.ClaimResponseViewSet, basename="ClaimResponse_R4")
    router.register(r'PractitionerRole', fhir_viewsets.PractitionerRoleViewSet, basename="PractitionerRole_R4")
    router.register(r'Practitioner', fhir_viewsets.PractitionerViewSet, basename="Practitioner_R4")
    router.register(r'CommunicationRequest', fhir_viewsets.CommunicationRequestViewSet,
                    basename="CommunicationRequest_R4")
    router.register(r'Communication', fhir_viewsets.CommunicationViewSet,
                    basename="Communication_R4")
    # code system for openimis medication
    router.register(r'CodeSystem/diagnosis', fhir_viewsets.CodeSystemOpenIMISDiagnosisViewSet,
                    basename="CodeSystem/diagnosis_R4")

# register endpoint for policy if used
if 'policy' in imis_modules:
    router.register(r'Coverage', fhir_viewsets.CoverageRequestQuerySet, basename="Coverage_R4")
    router.register(r'Contract', fhir_viewsets.ContractViewSet, basename="Contract_R4")

# register endpoint for policy holder if used
if 'policyholder' in imis_modules:
    router.register(r'Organization', fhir_viewsets.OrganisationViewSet, basename="Organisation_R4")
    router.register(r'CodeSystem/organization-ph-legal-form', fhir_viewsets.CodeSystemOrganizationPHLegalFormViewSet,
                    basename="CodeSystem/organization-ph-legal-form_R4")
    router.register(r'CodeSystem/organization-ph-activity', fhir_viewsets.CodeSystemOrganizationPHActivityViewSet,
                    basename="CodeSystem/organization-ph-activity_R4")

# register endpoint for policy holder if used
if 'invoice' in imis_modules:
    router.register(r'Invoice', fhir_viewsets.InvoiceViewSet, basename="Invoice_R4")

urlpatterns = [
    path('', include(router.urls)),
    path('docs/', SpectacularAPIView.as_view(), name='docs'),
    path('docs/swagger/', SpectacularSwaggerView.as_view(url_name='docs'), name='swagger-ui'),
    path('docs/redoc/', SpectacularRedocView.as_view(url_name='docs'), name='redoc'),
]

OrderedDict



from policy.services import EligibilityRequest
from api_fhir_r4.configurations import GeneralConfiguration, R4CoverageEligibilityConfiguration as Config
from api_fhir_r4.converters import BaseFHIRConverter, PatientConverter, ReferenceConverterMixin
from fhir.resources.money import Money
from fhir.resources.reference import Reference
from fhir.resources.period import Period
from fhir.resources.coverageeligibilityresponse import CoverageEligibilityResponse as FHIRCoverageEligibilityResponse, \
    CoverageEligibilityResponseInsuranceItem, \
    CoverageEligibilityResponseInsurance, CoverageEligibilityResponseInsuranceItemBenefit
from api_fhir_r4.models import CoverageEligibilityRequestV2 as FHIRCoverageEligibilityRequest
from api_fhir_r4.utils import TimeUtils
from insuree.models import InsureePolicy
from product.models import Product
from policy.services import EligibilityResponse
from django.db import connection
from claim.models import Claim, ClaimService, ClaimItem
from medical.models import Item, Service


class CoverageEligibilityRequestConverter(BaseFHIRConverter):

    @classmethod
    def to_fhir_obj(cls, coverage_eligibility_response, coverage_eligibility_request, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_eligibility_response = {}
        fhir_eligibility_response["status"] = 'active'
        fhir_eligibility_response["outcome"] = 'complete'

        reference_insurer = {}
        reference_insurer["reference"] = 'openIMIS'
        fhir_eligibility_response["insurer"] = reference_insurer

        reference_patient = {}
        reference_patient["reference"] = f'Patient/{coverage_eligibility_request.chf_id}'
        fhir_eligibility_response['patient'] = reference_patient

        reference_coverage_eligibility_request = {}
        reference_coverage_eligibility_request["reference"] = f'CoverageEligibilityRequest'
        fhir_eligibility_response['request'] = reference_coverage_eligibility_request

        fhir_eligibility_response["purpose"] = ["benefits"]
        fhir_eligibility_response["created"] = TimeUtils.date().isoformat()
        fhir_response = FHIRCoverageEligibilityResponse(**fhir_eligibility_response)
        for item in coverage_eligibility_response.items:
            if item.status in Config.get_fhir_active_policy_status():
                cls.build_fhir_insurance(fhir_response, item, coverage_eligibility_request)
        return fhir_response

    @classmethod
    def to_imis_obj(cls, fhir_coverage_eligibility_request, audit_user_id):
        fhir_coverage_eligibility_request = FHIRCoverageEligibilityRequest(**fhir_coverage_eligibility_request)
        chf_id = cls.build_imis_chf(fhir_coverage_eligibility_request)
        item_code, service_code = cls.build_imis_item_service(fhir_coverage_eligibility_request)
        return EligibilityRequest(chf_id, service_code, item_code)

    @classmethod
    def build_fhir_insurance(cls, fhir_response, item, request):
        result = CoverageEligibilityResponseInsurance.construct()
        cls.build_fhir_coverage(result, item.policy_uuid)
        cls.build_fhir_benefit_period(result, item.start_date, item.expiry_date)
        # get the data from SP
        try:
            with connection.cursor() as cur:
                sql = """\
                            DECLARE @MinDateService DATE, @MinDateItem DATE,
                                    @ServiceLeft INT, @ItemLeft INT,
                                    @isItemOK BIT, @isServiceOK BIT;
                            EXEC [dbo].[uspServiceItemEnquiry] @CHFID = %s, @ServiceCode = %s, @ItemCode = %s,
                                 @MinDateService = @MinDateService OUTPUT, @MinDateItem = @MinDateItem OUTPUT,
                                 @ServiceLeft = @ServiceLeft OUTPUT, @ItemLeft = @ItemLeft OUTPUT,
                                 @isItemOK = @isItemOK OUTPUT, @isServiceOK = @isServiceOK OUTPUT;
                            SELECT @MinDateService, @MinDateItem, @ServiceLeft, @ItemLeft, @isItemOK, @isServiceOK
                        """
                cur.execute(sql, (request.chf_id,
                                  request.service_code,
                                  request.item_code))
                res = cur.fetchone()  # retrieve the stored proc @Result table
                if res is None:
                    response_eligibility_sp = EligibilityResponse(
                        eligibility_request=None,
                        prod_id=None,
                        total_admissions_left=0,
                        total_visits_left=0,
                        total_consultations_left=0,
                        total_surgeries_left=0,
                        total_deliveries_left=0,
                        total_antenatal_left=0,
                        consultation_amount_left=0.0,
                        surgery_amount_left=0.0,
                        delivery_amount_left=0.0,
                        hospitalization_amount_left=0.0,
                        antenatal_amount_left=0.0,
                        min_date_service=None,
                        min_date_item=None,
                        service_left=0,
                        item_left=0,
                        is_item_ok=False,
                        is_service_ok=False
                    )

                (prod_id, total_admissions_left, total_visits_left, total_consultations_left, total_surgeries_left,
                 total_deliveries_left, total_antenatal_left, consultation_amount_left, surgery_amount_left,
                 delivery_amount_left,
                 hospitalization_amount_left, antenatal_amount_left) = res
                cur.nextset()
                (min_date_service, min_date_item, service_left,
                 item_left, is_item_ok, is_service_ok) = cur.fetchone()
                response_eligibility_sp = EligibilityResponse(
                    eligibility_request=request,
                    prod_id=prod_id or None,
                    total_admissions_left=total_admissions_left or 0,
                    total_visits_left=total_visits_left or 0,
                    total_consultations_left=total_consultations_left or 0,
                    total_surgeries_left=total_surgeries_left or 0,
                    total_deliveries_left=total_deliveries_left or 0,
                    total_antenatal_left=total_antenatal_left or 0,
                    consultation_amount_left=consultation_amount_left or 0.0,
                    surgery_amount_left=surgery_amount_left or 0.0,
                    delivery_amount_left=delivery_amount_left or 0.0,
                    hospitalization_amount_left=hospitalization_amount_left or 0.0,
                    antenatal_amount_left=antenatal_amount_left or 0.0,
                    min_date_service=min_date_service,
                    min_date_item=min_date_item,
                    service_left=service_left or 0,
                    item_left=item_left or 0,
                    is_item_ok=is_item_ok is True,
                    is_service_ok=is_service_ok is True
                )
        except:
            response_eligibility_sp = EligibilityResponse(
                eligibility_request=None,
                prod_id=None,
                total_admissions_left=0,
                total_visits_left=0,
                total_consultations_left=0,
                total_surgeries_left=0,
                total_deliveries_left=0,
                total_antenatal_left=0,
                consultation_amount_left=0.0,
                surgery_amount_left=0.0,
                delivery_amount_left=0.0,
                hospitalization_amount_left=0.0,
                antenatal_amount_left=0.0,
                min_date_service=None,
                min_date_item=None,
                service_left=0,
                item_left=0,
                is_item_ok=False,
                is_service_ok=False
            )

        # build coverag item - product/benefit
        result.item = []
        cls.build_fhir_benefit_item_element(result, response_eligibility_sp)
        # check services and items etc
        service_claim = ClaimService.objects.filter(
            claim__insuree__chf_id=request.chf_id,
            validity_to=None).values('service').distinct()
        item_claim = ClaimItem.objects.filter(
            claim__insuree__chf_id=request.chf_id,
            validity_to=None).values('item').distinct()
        # build coverag item - service
        if service_claim.count() > 0:
            cls.build_fhir_benefit_item_service_element(result, response_eligibility_sp, service_claim)
        # build coverag item - item
        if item_claim.count() > 0:
            cls.build_fhir_benefit_item_item_element(result, response_eligibility_sp, item_claim)
        if type(fhir_response.insurance) is not list:
            fhir_response.insurance = [result]
        else:
            fhir_response.insurance.append(result)

    @classmethod
    def build_fhir_coverage(cls, insurance, policy_uuid):
        reference_coverage = Reference.construct()
        reference_coverage.reference = F"Coverage/{policy_uuid}"
        insurance.coverage = reference_coverage

    @classmethod
    def build_fhir_benefit_period(cls, insurance, start_date, expiry_date):
        benefit_period = Period.construct()
        benefit_period.start = start_date
        benefit_period.end = expiry_date
        insurance.benefitPeriod = benefit_period

    @classmethod
    def build_fhir_benefit_item_element(cls, insurance, response):
        item = CoverageEligibilityResponseInsuranceItem.construct()
        system = F"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-category"
        item.category = cls.build_codeable_concept(
            system=system,
            code="benefit",
            display="Benefit Package"
        )
        cls.__build_item_product_name(fhir_item=item, prod_id=response.prod_id)
        item.benefit = []
        if response.total_admissions_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="admissions_left",
                display="total_admissions",
                value=response.total_admissions_left
            )
        if response.total_consultations_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="consultations_left",
                display="total_consultations",
                value=response.total_consultations_left
            )
        if response.total_surgeries_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="surgeries_left",
                display="total_surgeries",
                value=response.total_surgeries_left
            )
        if response.total_deliveries_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="deliveries_left",
                display="total_deliveries",
                value=response.total_deliveries_left
            )
        if response.total_visits_left:
            cls.build_fhir_int_item_benefit_element(
               item=item,
               code="visits_left",
               display="total_visits",
               value=response.total_visits_left
            )
        if response.total_antenatal_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="antenatal_left",
                display="total_antenatal",
                value=response.total_antenatal_left
            )
        if response.hospitalization_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="hospitalization_amount",
                display="hospitalization_amount",
                value=response.hospitalization_amount_left
            )
        if response.delivery_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="delivery_amount",
                display="delivery_amount",
                value=response.delivery_amount_left
            )
        if response.surgery_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="surgery_amount",
                display="surgery_amount",
                value=response.surgery_amount_left
            )
        if response.consultation_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="consultation_amount",
                display="consultation_amount",
                value=response.consultation_amount_left
            )
        if response.antenatal_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="antenatal_amount",
                display="antenatal_amount",
                value=response.antenatal_amount_left
            )
        insurance.item.append(item)

    @classmethod
    def build_fhir_benefit_item_item_element(cls, insurance, response, item_claim):
        item = CoverageEligibilityResponseInsuranceItem.construct()
        system = F"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-category"
        item.category = cls.build_codeable_concept(
            system=system,
            code="item",
            display="Item"
        )
        item.benefit = []
        if response.min_date_item:
            cls.build_fhir_string_item_benefit_element(
                item=item,
                code="min_date_item",
                display="Mininum date",
                value=response.min_date_item
            )
        if response.item_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="item_left",
                display="Items left",
                value=response.service_left
            )
        code = Item.objects.get(id=item_claim.first()['item'], validity_to=None).code
        item.productOrService = cls.build_simple_codeable_concept(code)
        item.excluded = not response.is_service_ok
        insurance.item.append(item)

    @classmethod
    def build_fhir_benefit_item_service_element(cls, insurance, response, service_claim):
        item = CoverageEligibilityResponseInsuranceItem.construct()
        system = F"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-category"
        item.category = cls.build_codeable_concept(
            system=system,
            code="service",
            display="Service"
        )
        item.benefit = []
        if response.min_date_service:
            cls.build_fhir_string_item_benefit_element(
                item=item,
                code="min_date_item",
                display="Mininum date",
                value=response.min_date_item
            )
        if response.item_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="service_left",
                display="Services left",
                value=response.item_left
            )
        code = Service.objects.get(id=service_claim.first()['service'], validity_to=None).code
        item.productOrService = cls.build_simple_codeable_concept(code)
        item.excluded = not response.is_item_ok
        insurance.item.append(item)

    @classmethod
    def build_fhir_int_item_benefit_element(cls, item, code, display, value):
        system = F"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-benefit-type"
        benefit = CoverageEligibilityResponseInsuranceItemBenefit.construct()
        benefit.type = cls.build_codeable_concept(
            system=system,
            code=code,
            display=display
        )
        benefit.allowedUnsignedInt = value
        item.benefit.append(benefit)

    @classmethod
    def build_fhir_money_item_benefit_element(cls, item, code, display, value):
        system = F"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-benefit-type"
        benefit = CoverageEligibilityResponseInsuranceItemBenefit.construct()
        benefit.type = cls.build_codeable_concept(
            system=system,
            code=code,
            display=display
        )
        money_value = Money.construct()
        money_value.value = value
        benefit.allowedMoney = money_value
        item.benefit.append(benefit)

    @classmethod
    def build_fhir_string_item_benefit_element(cls, item, code, display, value):
        system = F"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-benefit-type"
        benefit = CoverageEligibilityResponseInsuranceItemBenefit.construct()
        benefit.type = cls.build_codeable_concept(
            system=system,
            code=code,
            display=display
        )
        benefit.allowedString = value
        item.benefit.append(benefit)

    @classmethod
    def build_imis_chf(cls, fhir_coverage_eligibility_request):
        chf_id = None
        patient_reference = fhir_coverage_eligibility_request.patient
        if patient_reference:
            chf_id = PatientConverter.get_resource_id_from_reference(patient_reference)
        return chf_id

    @classmethod
    def build_imis_item_service(cls, fhir_coverage_eligibility_request):
        service_code = None
        item_code = None
        if fhir_coverage_eligibility_request.item:
            for item in fhir_coverage_eligibility_request.item:
                type_service = cls.__get_code_from_codeable_concept_by_coding_code(item.category)
                if type_service == 'item':
                    item_code = item.productOrService.text
                if type_service == 'service':
                    service_code = item.productOrService.text
        return item_code, service_code

    @classmethod
    def __get_code_from_codeable_concept_by_coding_code(cls, codeable_concept):
        service_code = None
        if codeable_concept:
            coding = cls.get_first_coding_from_codeable_concept(codeable_concept)
            if coding:
                service_code = coding.code
        return service_code

    @classmethod
    def __get_coverage_data(cls, response):
        policy = InsureePolicy.objects.filter(
            insuree__chf_id=response.eligibility_request.chf_id,
            validity_to__isnull=True).first().policy
        product = Product.objects.get(id=response.prod_id, validity_to__isnull=True)
        return policy, product

    @classmethod
    def __build_item_product_name(cls, fhir_item, prod_id):
        product_queryset = Product.objects.all().filter(id=prod_id, validity_to=None)
        product = product_queryset.first()
        fhir_item.name = product.name
        fhir_item.description = product.code



from api_fhir_r4.configurations import GeneralConfiguration, R4CoverageConfig
from django.utils.translation import gettext as _
from policy.models import Policy


class ContractStatus(object):

    contract_status = {
        f"{Policy.STATUS_ACTIVE}": R4CoverageConfig.get_status_policy_code(),
        f"{Policy.STATUS_IDLE}": R4CoverageConfig.get_status_offered_code(),
        f"{Policy.STATUS_EXPIRED}": R4CoverageConfig.get_status_terminated_code(),
        f"{Policy.STATUS_SUSPENDED}": R4CoverageConfig.get_status_disputed_code(),
    }

    @classmethod
    def imis_map_status(cls, code, imis_policy):
        status = {
            R4CoverageConfig.get_status_idle_code(): imis_policy.STATUS_IDLE,
            R4CoverageConfig.get_status_active_code(): imis_policy.STATUS_ACTIVE,
            R4CoverageConfig.get_status_suspended_code(): imis_policy.STATUS_SUSPENDED,
            R4CoverageConfig.get_status_expired_code(): imis_policy.STATUS_EXPIRED,
        }
        return status[code]


class ContractState(object):

    contract_state = {
        f"{Policy.STAGE_NEW}": R4CoverageConfig.get_status_offered_code(),
        f"{Policy.STAGE_RENEWED}": R4CoverageConfig.get_status_renewed_code(),
    }

    @classmethod
    def imis_map_stage(cls, code, imis_policy):
        codes = {
            R4CoverageConfig.get_status_offered_code(): imis_policy.STAGE_NEW,
            R4CoverageConfig.get_status_active_code(): imis_policy.STAGE_RENEWED
        }
        return codes[code]


class PayTypeMapping(object):

    pay_type = {
        "B": _("Bank transfer"),
        "C": _("Cash"),
        "M": _("Mobile phone"),
        "F": _("Funding")
    }



from fhir.resources.invoice import Invoice

from api_fhir_r4.converters import InvoiceConverter
from api_fhir_r4.tests.mixin import ConvertJsonToFhirTestMixin
from api_fhir_r4.tests.mixin.invoiceTestMixin import InvoiceTestMixin
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin


class InvoiceConverterTestCase(InvoiceTestMixin, LogInMixin, ConvertJsonToFhirTestMixin):
    _TEST_USER_NAME = "TestUserTest2"

    converter = InvoiceConverter
    fhir_resource = Invoice
    json_repr = 'test/test_invoice.json'

    def test_to_fhir_obj(self):
        user = self.get_or_create_user_api()
        imis_invoice, imis_invoice_line_item = self.create_test_imis_instance()
        imis_invoice.save(username=user.username)
        imis_invoice_line_item.save(username=user.username)
        fhir_invoice = self.converter.to_fhir_obj(imis_invoice)
        self.verify_fhir_instance(fhir_invoice)

    def test_to_imis_obj(self):
        self.assertRaises(NotImplementedError, InvoiceConverter.to_imis_obj, object(), 1)



import logging

from django.core.exceptions import MultipleObjectsReturned
from location.models import HealthFacility, Location, HealthFacilityLegalForm
from claim.models import ClaimAdmin
from fhir.resources.address import Address

from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin, LocationConverter, PersonConverterMixin
from fhir.resources.organization import Organization
from fhir.resources.organization import OrganizationContact
from fhir.resources.extension import Extension
from api_fhir_r4.mapping.organizationMapping import HealthFacilityOrganizationTypeMapping
from api_fhir_r4.models.imisModelEnums import ImisLocationType
from api_fhir_r4.utils import DbManagerUtils
from django.utils.translation import gettext as _

logger = logging.getLogger(__name__)


class HealthFacilityOrganisationConverter(BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_organisation: HealthFacility, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_organisation = Organization()
        cls.build_fhir_pk(fhir_organisation, imis_organisation, reference_type)
        cls.build_fhir_extensions(fhir_organisation, imis_organisation)
        cls.build_fhir_identifiers(fhir_organisation, imis_organisation)
        cls.build_fhir_type(fhir_organisation, imis_organisation)
        cls.build_fhir_name(fhir_organisation, imis_organisation)
        cls.build_fhir_telecom(fhir_organisation, imis_organisation)
        cls.build_hf_address(fhir_organisation, imis_organisation, reference_type)
        cls.build_contacts(fhir_organisation, imis_organisation)
        return fhir_organisation

    @classmethod
    def to_imis_obj(cls, fhir_organisation, audit_user_id):
        errors = []
        fhir_hf = Organization(**fhir_organisation)
        imis_hf = HealthFacility()
        imis_hf.audit_user_id = audit_user_id
        cls.build_imis_hf_identiftier(imis_hf, fhir_hf, errors)
        cls.build_imis_hf_name(imis_hf, fhir_hf, errors)
        cls.build_imis_hf_telecom(imis_hf, fhir_hf, errors)
        cls.build_imis_hf_address(imis_hf, fhir_hf, errors)
        cls.build_imis_hf_level(imis_hf, fhir_hf, errors)
        cls.build_imis_care_type(imis_hf, fhir_hf, errors)
        cls.build_imis_legal_form(imis_hf, fhir_hf, errors)
        cls.build_imis_parent_location_id(imis_hf, fhir_hf, errors)
        cls.check_errors(errors)
        return imis_hf

    @classmethod
    def get_fhir_resource_type(cls):
        return Organization

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        healthfacility_uuid = cls.get_resource_id_from_reference(reference).lower()
        return DbManagerUtils.get_object_or_none(HealthFacility, uuid=healthfacility_uuid)

    @classmethod
    def build_fhir_extensions(cls, fhir_organisation: Organization, imis_organisation: HealthFacility):
        extensions = [
            cls.__legal_form_extension(imis_organisation),
            cls.__level_extension(imis_organisation),
            cls.__type_extension(imis_organisation)
        ]
        fhir_organisation.extension = [ext for ext in extensions if ext]

    @classmethod
    def build_fhir_identifiers(cls, fhir_organisation, imis_organisation):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_organisation)
        fhir_organisation.identifier = identifiers

    @classmethod
    def build_fhir_type(cls, fhir_organisation, imis_organisation):
        organisation_type = cls.build_codeable_concept(code=HealthFacilityOrganizationTypeMapping.ORGANIZATION_TYPE)
        fhir_organisation.type = [organisation_type]

    @classmethod
    def build_fhir_name(cls, fhir_organisation: Organization, imis_organisation: HealthFacility):
        name = imis_organisation.name
        fhir_organisation.name = name

    @classmethod
    def build_fhir_telecom(cls, fhir_organisation: Organization, imis_organisation: HealthFacility):
        telecom = []

        if imis_organisation.email:
            telecom.append(cls._build_email_contact_point(imis_organisation))

        if imis_organisation.phone:
            telecom.append(cls._build_phone_contact_point(imis_organisation))

        if imis_organisation.fax:
            telecom.append(cls._build_fax_contact_point(imis_organisation))

        fhir_organisation.telecom = telecom

    @classmethod
    def _build_email_contact_point(cls, imis_organisation):
        return cls.build_fhir_contact_point(
            value=imis_organisation.email,
            system=HealthFacilityOrganizationTypeMapping.EMAIL_CONTACT_POINT_SYSTEM
        )

    @classmethod
    def _build_phone_contact_point(cls, imis_organisation):
        return cls.build_fhir_contact_point(
            value=imis_organisation.phone,
            system=HealthFacilityOrganizationTypeMapping.PHONE_CONTACT_POINT_SYSTEM
        )

    @classmethod
    def _build_fax_contact_point(cls, imis_organisation):
        return cls.build_fhir_contact_point(
            value=imis_organisation.fax,
            system=HealthFacilityOrganizationTypeMapping.FAX_CONTACT_POINT_SYSTEM
        )

    @classmethod
    def build_hf_address(cls, fhir_organisation: Organization, imis_organisation: HealthFacility, reference_type):
        address = Address.construct()
        if imis_organisation.address:
            address.line = [imis_organisation.address]

        # Hospitals are expected to be on district level
        address.district = imis_organisation.location.name
        address.state = imis_organisation.location.parent.name
        address.type = 'physical'
        address.extension = [cls._build_address_ext(imis_organisation, reference_type)]

        fhir_organisation.address = [address]

    @classmethod
    def _build_address_ext(cls, imis_organisation, reference_type):
        address_ref = LocationConverter\
            .build_fhir_resource_reference(imis_organisation, 'Organization', reference_type=reference_type)
        address_ext = Extension.construct()
        address_ext.url = HealthFacilityOrganizationTypeMapping.ADDRESS_LOCATION_REFERENCE_URL
        address_ext.valueReference = address_ref
        return address_ext

    @classmethod
    def build_contacts(cls, fhir_organisation, imis_organisation):
        contracts = []
        relevant_claim_admins = ClaimAdmin.objects\
            .filter(health_facility=imis_organisation, validity_to__isnull=True).distinct()

        for admin in relevant_claim_admins:
            contracts.append(
                cls._build_claim_admin_contract(admin)
            )

        fhir_organisation.contact = contracts

    @classmethod
    def _build_claim_admin_contract(cls, admin):
        contract = OrganizationContact.construct()
        contract.purpose = cls.build_codeable_concept(**HealthFacilityOrganizationTypeMapping.CONTRACT_PURPOSE)
        name = cls.build_fhir_names_for_person(admin)
        contract.name = name
        return contract

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_location_code_type()

    @classmethod
    def get_reference_obj_uuid(cls, imis_organisation):
        return imis_organisation.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_organisation):
        return imis_organisation.id

    @classmethod
    def get_reference_obj_code(cls, imis_organisation):
        return imis_organisation.code

    @classmethod
    def __legal_form_extension(cls, imis_organisation: HealthFacility):
        if imis_organisation.legal_form:
            legal_form = imis_organisation.legal_form.code
            extension = Extension.construct()
            extension.url = f'{GeneralConfiguration.get_system_base_url()}StructureDefinition/organization-legal-form'
            extension.valueCodeableConcept = cls.build_codeable_concept(
                code=legal_form,
                system=HealthFacilityOrganizationTypeMapping.LEGAL_FORM_SYSTEM,
                display=HealthFacilityOrganizationTypeMapping.LEGAL_FORM_MAPPING.get(legal_form)
            )
            return extension

    @classmethod
    def __level_extension(cls, imis_organisation):
        if not imis_organisation.level:
            return

        level = imis_organisation.level

        level_display = HealthFacilityOrganizationTypeMapping.LEVEL_DISPLAY_MAPPING.get(level, None)
        if not level_display:
            logger.warning(f'Failed to build level display for HF Level {level}.')

        extension = Extension.construct()
        extension.url = f'{GeneralConfiguration.get_system_base_url()}/StructureDefinition/organization-hf-level'
        extension.valueCodeableConcept = cls.build_codeable_concept(
            code=level,
            system=HealthFacilityOrganizationTypeMapping.LEVEL_SYSTEM,
            display=level_display
        )

        return extension
    @classmethod
    def __type_extension(cls, imis_organisation: HealthFacility):
        if imis_organisation.care_type and imis_organisation.care_type != ' ':
            care_type = imis_organisation.care_type
            care_type_display = HealthFacilityOrganizationTypeMapping.TYPE_DISPLAY_MAPPING.get(care_type, None)

            extension = Extension.construct()
            extension.url = f'{GeneralConfiguration.get_system_base_url()}StructureDefinition/organization-hf-care-type'
            extension.valueCodeableConcept = cls.build_codeable_concept(
                code=care_type,
                system=HealthFacilityOrganizationTypeMapping.TYPE_SYSTEM,
                display=care_type_display
            )
            return extension

    @classmethod
    def build_imis_hf_identiftier(cls, imis_hf, fhir_hf, errors):
        value = cls.get_fhir_identifier_by_code(
            fhir_hf.identifier,
            R4IdentifierConfig.get_fhir_generic_type_code()
        )

        if value:
            imis_hf.code = value
        cls.valid_condition(imis_hf.code is None, _('Missing HF Organization code'), errors)

    @classmethod
    def build_imis_hf_name(cls, imis_hf, fhir_hf, errors):
        imis_hf.name = fhir_hf.name
        cls.valid_condition(imis_hf.code is None, _('Missing HF Name'), errors)

    @classmethod
    def build_imis_hf_telecom(cls, imis_hf, fhir_hf, errors):
        cls._build_imis_hf_email(imis_hf, fhir_hf, errors)
        cls._build_imis_hf_fax(imis_hf, fhir_hf, errors)
        cls._build_imis_hf_phone(imis_hf, fhir_hf, errors)

    @classmethod
    def _build_imis_hf_email(cls, imis_hf, fhir_hf, errors):
        imis_hf.email = cls.__get_unique_telecom(
            fhir_hf,
            HealthFacilityOrganizationTypeMapping.EMAIL_CONTACT_POINT_SYSTEM,
            errors
        )

    @classmethod
    def _build_imis_hf_fax(cls, imis_hf, fhir_hf, errors):
        imis_hf.fax = cls.__get_unique_telecom(
            fhir_hf,
            HealthFacilityOrganizationTypeMapping.FAX_CONTACT_POINT_SYSTEM,
            errors
        )

    @classmethod
    def _build_imis_hf_phone(cls, imis_hf, fhir_hf, errors):
        imis_hf.phone = cls.__get_unique_telecom(
            fhir_hf,
            HealthFacilityOrganizationTypeMapping.PHONE_CONTACT_POINT_SYSTEM,
            errors
        )

    @classmethod
    def __get_unique_telecom(cls, fhir_hf, system, errors):
        if not fhir_hf.telecom:
            return None

        telecom = cls.get_contract_points_by_system_code(
            fhir_hf.telecom,
            HealthFacilityOrganizationTypeMapping.FAX_CONTACT_POINT_SYSTEM)

        cls.valid_condition(
            len(telecom) > 1,
            _(f'More than one contact point of system {system} assigned to health facility'),
            errors
        )
        return telecom[0] if telecom else None

    @classmethod
    def build_imis_hf_address(cls, imis_hf: HealthFacility, fhir_hf: Organization, errors):
        if fhir_hf.address:
            imis_hf.address = fhir_hf.address[0].line[0]

    @classmethod
    def build_imis_hf_level(cls, imis_hf, fhir_hf, errors):
        ext_url_suffix = 'organization-hf-level'
        imis_hf.level = cls.__get_extension_by_url_suffix(fhir_hf.extension, ext_url_suffix)

    @classmethod
    def build_imis_care_type(cls, imis_hf, fhir_hf, errors):
        ext_url_suffix = 'organization-hf-care-type'
        value = cls.__get_extension_by_url_suffix(fhir_hf.extension, ext_url_suffix)
        if value:
            imis_hf.care_type = value
        else:
            cls.valid_condition(True, _("Extension with HF care type not found"), errors)

    @classmethod
    def build_imis_legal_form(cls, imis_hf, fhir_hf, errors):
        ext_url_suffix = 'organization-legal-form'
        value = cls.__get_extension_by_url_suffix(fhir_hf.extension, ext_url_suffix)
        cls.valid_condition(
            value not in HealthFacilityOrganizationTypeMapping.LEGAL_FORM_MAPPING.keys(),
            _("Invalid HF legal form code, has to be one of %(codes)s") % {
                'codes': HealthFacilityOrganizationTypeMapping.LEGAL_FORM_MAPPING.keys()},
            errors)
        if value:
            legal_form = HealthFacilityLegalForm.objects.get(code=value)
            imis_hf.legal_form = legal_form
        else:
            cls.valid_condition(True, _("Extension with HF legal form not found"), errors)

    @classmethod
    def __get_extension_by_url_suffix(cls, extensions, ext_url_suffix):
        level_ext = next((x for x in extensions if x.url.endswith(ext_url_suffix)), None)
        if level_ext:
            return cls.get_first_coding_from_codeable_concept(
                level_ext.valueCodeableConcept).code
        else:
            return None

    @classmethod
    def build_imis_parent_location_id(cls, imis_hf, fhir_hf, errors):
        address = fhir_hf.address[0]
        matching_locations = Location.objects \
            .filter(
                validity_to__isnull=True,
                name=address.district,
                parent__name=address.state,
                type="D"  # HF is expected to be at district level
            ).distinct()\
            .all()

        if matching_locations.count() != 1:
            msg = cls.__get_invalid_location_msg(address, matching_locations)
            cls.valid_condition(matching_locations.count() == 1, msg, errors)
            return
        else:
            imis_hf.location = matching_locations.first()

    @classmethod
    def __get_invalid_location_msg(cls, address, matching_locations):
        count = matching_locations.count()
        if count == 0:
            return _(F"No matching location for district {address.district}, state {address.state}.")
        elif count > 1:
            return _(F"More than one matching location district {address.district}, state {address.state}:\n"
                     F"{matching_locations}.")



import logging
from abc import abstractmethod, ABC
from typing import List

from django.core.exceptions import ObjectDoesNotExist, FieldError
from django.http import Http404

from rest_framework import mixins

from api_fhir_r4.model_retrievers import GenericModelRetriever
from rest_framework.response import Response

from api_fhir_r4.multiserializer.mixins import MultiSerializerUpdateModelMixin, MultiSerializerRetrieveModelMixin

logger = logging.getLogger(__name__)


class GenericMultiIdentifierMixin(ABC):
    lookup_field = 'identifier'

    @property
    @abstractmethod
    def retrievers(self) -> List[GenericModelRetriever]:
        # Identifiers available for given resource
        pass

    def _get_object_with_first_valid_retriever(self, identifier):
        for retriever in self.retrievers:
            if retriever.identifier_validator(identifier):
                try:
                    queryset = retriever.retriever_additional_queryset_filtering(self.get_queryset())
                    resource = retriever.get_model_object(queryset, identifier)

                    # May raise a permission denied
                    self.check_object_permissions(self.request, resource)
                    return retriever.serializer_reference_type, resource
                except ObjectDoesNotExist as e:
                    logger.exception(
                        F"Failed to retrieve object from queryset {self.get_queryset()} using"
                        F"identifier {identifier} for matching retriever: {retriever}"
                    )

        # Raise Http404 if resource couldn't be fetched with any of the retrievers
        raise Http404(f"Resource for identifier {identifier} not found")


class MultiIdentifierRetrieverMixin(mixins.RetrieveModelMixin, GenericMultiIdentifierMixin, ABC):

    def retrieve(self, request, *args, **kwargs):
        ref_type, instance = self._get_object_with_first_valid_retriever(kwargs['identifier'])
        serializer = self.get_serializer(instance, reference_type=ref_type)
        return Response(serializer.data)


class MultiIdentifierUpdateMixin(mixins.UpdateModelMixin, GenericMultiIdentifierMixin, ABC):

    def update(self, request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        ref_type, instance = self._get_object_with_first_valid_retriever(kwargs['identifier'])
        serializer = self.get_serializer(instance, data=request.data, partial=partial, reference_type=ref_type)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        if getattr(instance, '_prefetched_objects_cache', None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        return Response(serializer.data)


class GenericMultiIdentifierForManySerializers(GenericMultiIdentifierMixin, ABC):

    def _get_object_with_first_valid_retriever(self, queryset, identifier):
        for retriever in self.retrievers:
            if retriever.identifier_validator(identifier):
                try:
                    queryset = retriever.retriever_additional_queryset_filtering(queryset)
                    resource = retriever.get_model_object(queryset, identifier)

                    # May raise a permission denied
                    self.check_object_permissions(self.request, resource)
                    return retriever.serializer_reference_type, resource
                except ObjectDoesNotExist:
                    logger.exception(
                        F"Failed to retrieve object from queryset {queryset} using"
                        F"identifier {identifier} for matching retriever: {retriever}"
                    )
                except FieldError:
                    logger.exception(F"Failed to retrieve object from queryset {queryset} using"
                                     F"{self.lookup_field}, field does not available for given model {queryset.model}")
        return None, None


class MultiIdentifierUpdateManySerializersMixin(MultiSerializerUpdateModelMixin,
                                                GenericMultiIdentifierForManySerializers, ABC):
    def update(self, request, *args, **kwargs):
        self._validate_update_request()
        partial = kwargs.pop('partial', False)
        results = []
        for serializer, (qs, _, _) in self.get_eligible_serializers_iterator():
            ref_type, instance = self._get_object_with_first_valid_retriever(qs, kwargs['identifier'])
            update_result = self._update_for_serializer(serializer, instance, request.data, partial,
                                                        reference_type=ref_type)
            results.append(update_result)

        response = results[0]  # By default there should be only one eligible serializer
        return Response(response)


class MultiIdentifierRetrieveManySerializersMixin(MultiSerializerRetrieveModelMixin,
                                                  GenericMultiIdentifierForManySerializers, ABC):
    def retrieve(self, request, *args, **kwargs):
        self._validate_retrieve_model_request()
        retrieved = []
        for serializer, (qs, _, _) in self.get_eligible_serializers_iterator():
            ref_type, instance = self._get_object_with_first_valid_retriever(qs, kwargs['identifier'])
            if instance:
                serializer = serializer(instance, reference_type=ref_type)
                if serializer.data:
                    retrieved.append(serializer.data)

        if len(retrieved) > 1:
            raise ValueError("Ambiguous retrieve result, object found for multiple serializers.")
        if len(retrieved) == 0:
            raise Http404(f"Resource for identifier {kwargs['identifier']} not found")

        return Response(retrieved[0])



from api_fhir_r4.configurations import BaseConfiguration, GeneralConfiguration, R4ApiFhirConfig
from django.conf import settings


class ModuleConfiguration(BaseConfiguration):

    __REST_FRAMEWORK = {
        'EXCEPTION_HANDLER': 'api_fhir_r4.exceptions.fhir_api_exception_handler'
    }

    @classmethod
    def build_configuration(cls, cfg):
        GeneralConfiguration.build_configuration(cfg)
        cls.get_r4().build_configuration(cfg)
        cls.configure_api_error_handler()

    @classmethod
    def get_r4(cls):
        return R4ApiFhirConfig

    @classmethod
    def configure_api_error_handler(cls):
        config = cls.get_config()
        rest_settings = settings.__getattr__("REST_FRAMEWORK")
        config.default_api_error_handler = rest_settings.get("EXCEPTION_HANDLER")
        rest_settings.update(cls.__REST_FRAMEWORK)

    @classmethod
    def get_default_api_error_handler(cls):
        return cls.get_config().default_api_error_handler



from api_fhir_r4.converters.coverageConverter import CoverageConverter
from api_fhir_r4.serializers import BaseFHIRSerializer
from api_fhir_r4.models import CoverageV2, CoverageClassV2
from rest_framework import serializers
from policy.models import Policy
import copy
from api_fhir_r4.exceptions import FHIRException

class CoverageSerializer(BaseFHIRSerializer):
    fhirConverter = CoverageConverter

    def create(self, validated_data):
        family = validated_data.get('family_id')
        if Policy.objects.filter(family_id=family).count() > 0:
            raise FHIRException('Exists coverage with the family  provided')
        copied_data = copy.deepcopy(validated_data)
        del copied_data['_state']
        return Policy.objects.create(**copied_data)
        
    def update(self, instance, validated_data):
        instance.status = validated_data.get('status', instance.status)
        instance.value = validated_data.get('value', instance.value)
        instance.save()
        return instance


from api_fhir_r4.converters import GroupConverter

from fhir.resources.group import Group
from api_fhir_r4.tests import GroupTestMixin
from api_fhir_r4.tests.mixin import ConvertToImisTestMixin, ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class GroupConverterTestCase(GroupTestMixin,
                             ConvertToImisTestMixin,
                             ConvertToFhirTestMixin,
                             ConvertJsonToFhirTestMixin):
    converter = GroupConverter
    fhir_resource = Group
    json_repr = 'test/test_group.json'



from rest_framework.views import APIView

from api_fhir_r4.multiserializer import MultiSerializerSerializerClass
from api_fhir_r4.paginations import FhirBundleResultsSetPagination
from api_fhir_r4.permissions import FHIRApiPermissions
from api_fhir_r4.views import CsrfExemptSessionAuthentication


class BaseFHIRView(APIView):
    pagination_class = FhirBundleResultsSetPagination
    permission_classes = (FHIRApiPermissions,)
    authentication_classes = [CsrfExemptSessionAuthentication] + APIView.settings.DEFAULT_AUTHENTICATION_CLASSES


class BaseMultiserializerFHIRView(BaseFHIRView):
    serializer_class = MultiSerializerSerializerClass



from api_fhir_r4.configurations import BaseConfiguration
from api_fhir_r4.defaultConfig import DEFAULT_CFG
from django.conf import settings


class GeneralConfiguration(BaseConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        config = cls.get_config()
        config.default_audit_user_id = cfg['default_audit_user_id']
        config.gender_codes = cfg['gender_codes']
        config.base_url = cfg['base_url']
        config.default_value_of_patient_head_attribute = cfg['default_value_of_patient_head_attribute']
        config.default_value_of_patient_card_issued_attribute = cfg['default_value_of_patient_card_issued_attribute']
        config.default_value_of_location_offline_attribute = cfg['default_value_of_location_offline_attribute']
        config.default_value_of_location_care_type = cfg['default_value_of_location_care_type']
        config.default_response_page_size = cfg['default_response_page_size']
        config.claim_rule_engine_validation = cfg['claim_rule_engine_validation']

    @classmethod
    def get_default_audit_user_id(cls):
        return cls.get_config_attribute("default_audit_user_id")

    @classmethod
    def get_male_gender_code(cls):
        return cls.get_config_attribute("gender_codes").get('male', 'M')

    @classmethod
    def get_female_gender_code(cls):
        return cls.get_config_attribute("gender_codes").get('female', 'F')

    @classmethod
    def get_other_gender_code(cls):
        return cls.get_config_attribute("gender_codes").get('other', 'O')

    @classmethod
    def get_default_value_of_patient_head_attribute(cls):
        return cls.get_config_attribute("default_value_of_patient_head_attribute")

    @classmethod
    def get_default_value_of_patient_card_issued_attribute(cls):
        return cls.get_config_attribute("default_value_of_patient_card_issued_attribute")

    @classmethod
    def get_default_value_of_location_offline_attribute(cls):
        return cls.get_config_attribute("default_value_of_location_offline_attribute")

    @classmethod
    def get_default_value_of_location_care_type(cls):
        return cls.get_config_attribute("default_value_of_location_care_type")

    @classmethod
    def get_default_response_page_size(cls):
        return cls.get_config_attribute("default_response_page_size")

    @classmethod
    def get_claim_rule_engine_validation(cls):
        return cls.get_config_attribute("claim_rule_engine_validation")

    @classmethod
    def show_system(cls):
        return 1

    @classmethod
    def get_system_base_url(cls):
        return cls.get_config_attribute("base_url")

    @classmethod
    def get_host_domain(cls):
        url = cls.get_base_url()
        if url.startswith('/'):
            return f'http://{settings.SITE_URL()}'
        else:
            return ''

    @classmethod        
    def get_base_url(cls):
        MODULE_NAME = 'api_fhir_r4'
        site_root = settings.SITE_ROOT()
        if site_root is not None:
            base_url = '/' + site_root
        if base_url.endswith('/'):
            return base_url + MODULE_NAME + '/'
        else:
            return base_url + '/'+MODULE_NAME+'/'



from api_fhir_r4.converters.coverageConverter import CoverageConverter
from api_fhir_r4.tests import CoverageTestMixin
from api_fhir_r4.models import CoverageV2 as Coverage
from api_fhir_r4.tests.mixin import ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class CoverageConverterTestCase(CoverageTestMixin,
                                ConvertToFhirTestMixin,
                                ConvertJsonToFhirTestMixin):
    converter = CoverageConverter
    fhir_resource = Coverage
    json_repr = 'test/test_coverage.json'
